"""
Methods that execute specific queries against the SQLite database for the CONDITIONS table.
This supports the policy_sentry query functionality
"""

from __future__ import annotations

import functools
import logging
from typing import cast

from policy_sentry.querying.actions import get_action_data
from policy_sentry.shared.iam_data import get_service_prefix_data
from policy_sentry.util.conditions import is_condition_key_match

logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=1024)
def get_condition_keys_for_service(service_prefix: str) -> list[str]:
    """
    Get a list of available conditions per AWS service

    Arguments:
        service_prefix: An AWS service prefix, like `s3` or `kms`
    Returns:
        List: A list of condition keys
    """
    service_prefix_data = get_service_prefix_data(service_prefix)
    return list(service_prefix_data["conditions"])


# Per condition key name
# pylint: disable=inconsistent-return-statements
def get_condition_key_details(service_prefix: str, condition_key_name: str) -> dict[str, str]:
    """
    Get details about a specific condition key in JSON format

    Arguments:
        service_prefix: An AWS service prefix, like `ec2` or `kms`
        condition_key_name: The name of a condition key, like `ec2:Vpc`
    Returns:
        Dictionary: Metadata about the condition key
    """
    service_prefix_data = get_service_prefix_data(service_prefix)
    for condition_name, condition_data in service_prefix_data["conditions"].items():
        if is_condition_key_match(condition_name, condition_key_name):
            return {
                "name": condition_name,
                "description": condition_data["description"],
                "condition_value_type": condition_data["type"].lower(),
            }

    return {}


def get_conditions_for_action_and_raw_arn(action: str, raw_arn: str) -> list[str]:
    """
    Get a list of conditions available to an action.

    Arguments:
        action: The IAM action, like s3:GetObject
        raw_arn: The raw ARN format specific to the action
    Returns:
        List: A list of condition keys
    """
    service_prefix, action_name = action.split(":")
    action_data = get_action_data(service_prefix, action_name)
    conditions = []
    for action_info in action_data[service_prefix]:
        if action_info["resource_arn_format"].lower() == raw_arn.lower():
            conditions.extend(action_info["condition_keys"])
    return conditions


def get_condition_keys_available_to_raw_arn(raw_arn: str) -> list[str]:
    """
    Get a list of condition keys available to a RAW ARN

    Arguments:
        raw_arn: The value in the database, like arn:${Partition}:s3:::${BucketName}/${ObjectName}
    Returns:
        List: A list of condition keys
    """
    results = set()
    elements = raw_arn.split(":", 5)
    service_prefix = elements[2]
    service_prefix_data = get_service_prefix_data(service_prefix)
    for resource_data in service_prefix_data["resources"].values():
        if resource_data["arn"] == raw_arn:
            results.update(resource_data["condition_keys"])

    return list(results)


# pylint: disable=inconsistent-return-statements
def get_condition_value_type(condition_key: str) -> str | None:
    """
    Get the data type of the condition key - like Date, String, etc.

    Arguments:
        condition_key: A condition key, like a4b:filters_deviceType
    Returns:
        String: type of the condition key, like Bool, Date, String, etc.
    """
    service_prefix, _ = condition_key.split(":")
    service_prefix_data = get_service_prefix_data(service_prefix)

    for condition_key_entry, condition_key_data in service_prefix_data["conditions"].items():
        if is_condition_key_match(condition_key_entry, condition_key):
            return cast("str", condition_key_data["type"]).lower()

    return None



import json
from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

IAM_DATA_PATH = Path("policy_sentry/shared/data/iam-definition.json")


class MinifyIAMDataHook(BuildHookInterface):
    """Hatch build hook that minifies the IAM data store before building."""

    PLUGIN_NAME = "minify_iam_data"

    def initialize(self, _version, _build_data):
        dest = Path(self.root) / IAM_DATA_PATH

        dest.parent.mkdir(parents=True, exist_ok=True)
        minified = json.dumps(json.loads(IAM_DATA_PATH.read_bytes()), separators=(",", ":"))
        dest.write_text(minified)

    def finalize(self, _version, _build_data, _artifact_path):
        # after the build, pretty format the json file again
        file_path = Path(self.root) / IAM_DATA_PATH
        file_path.write_text(json.dumps(json.loads(file_path.read_bytes()), indent=2))



import unittest
import json
from policy_sentry.writing.sid_group import SidGroup
from policy_sentry.writing.minimize import minimize_statement_actions
from policy_sentry.querying.all import get_all_actions
from policy_sentry.util.policy_files import get_sid_names_from_policy


class MinimizeWildcardActionsTestCase(unittest.TestCase):
    def test_minimize_statement_actions(self):
        actions_to_minimize = [
            "kms:CreateGrant",
            "kms:CreateCustomKeyStore",
            "ec2:AuthorizeSecurityGroupEgress",
            "ec2:AuthorizeSecurityGroupIngress",
        ]
        desired_result = ["ec2:authorizes*", "kms:createc*", "kms:createg*"]
        all_actions = get_all_actions(lowercase=True)
        minchars = None
        self.maxDiff = None
        # minimized_actions_list = minimize_statement_actions(desired_actions, all_actions, minchars)
        self.assertListEqual(
            sorted(minimize_statement_actions(actions_to_minimize, all_actions, minchars)),
            sorted(desired_result),
        )

    def test_minimize_statement_actions_funky_case(self):
        actions_to_minimize = [
            "kms:creategrant",
            "kms:createcustomkeystore",
            "ec2:authorizesecuritygroupegress",
            "ec2:authorizesecuritygroupingress",
        ]
        desired_result = ["ec2:authorizes*", "kms:createc*", "kms:createg*"]
        all_actions = get_all_actions(lowercase=True)
        minchars = None
        self.maxDiff = None
        # minimized_actions_list = minimize_statement_actions(desired_actions, all_actions, minchars)
        self.assertListEqual(
            sorted(minimize_statement_actions(actions_to_minimize, all_actions, minchars)),
            sorted(desired_result),
        )

    def test_minimize_statement_actions_with_min_chars(self):
        actions_to_minimize = [
            "ec2:AuthorizeSecurityGroupEgress",
            "ec2:AuthorizeSecurityGroupIngress",
        ]
        desired_result = ["ec2:authorizesecurity*"]
        all_actions = get_all_actions(lowercase=True)
        minchars = 17
        self.maxDiff = None
        self.assertListEqual(
            sorted(minimize_statement_actions(actions_to_minimize, all_actions, minchars)),
            sorted(desired_result),
        )

    def test_minimize_statement_actions_with_wildcard(self):
        actions_to_minimize = ["ec2:*"]
        desired_result = ["ec2:*"]
        all_actions = get_all_actions(lowercase=True)
        minchars = None
        self.maxDiff = None
        self.assertListEqual(
            sorted(minimize_statement_actions(actions_to_minimize, all_actions, minchars)),
            sorted(desired_result),
        )

    def test_minimize_rw_same_one(self):
        cfg = {
            "mode": "crud",
            "name": "",
            "read": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
            ],
            "write": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
            ],
        }
        sid_grp = SidGroup()
        write_format = sid_grp.process_template(cfg, minimize=0)

        """
        Expected result:
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "SsmMultParametermyparameter",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:getpar*",
                        "ssm:deletepar*",
                        "ssm:la*",
                        "ssm:putp*"
                    ],
                    "Resource": [
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter"
                    ]
                }
            ]
        }
        """
        # To future-proof this unit test...
        # (1) check that there is only one SID so it was combined during minimization
        sid_names = get_sid_names_from_policy(write_format)
        self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

    def test_minimize_rw_same(self):
        cfg = {
            "mode": "crud",
            "name": "",
            "read": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter2",
            ],
            "write": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter2",
            ],
        }
        sid_grp = SidGroup()
        write_format = sid_grp.process_template(cfg, minimize=1)
        # print(json.dumps(write_format, indent=4))
        """
        the output will look like:

        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "SsmParameterMyparameter",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:getpar*",
                        "ssm:deletepar*",
                        "ssm:la*",
                        "ssm:putp*"
                    ],
                    "Resource": [
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter2"
                    ]
                }
            ]
        }
        """
        # To future-proof this unit test...
        # (1) check that there is only one SID so it was combined during minimization
        sid_names = get_sid_names_from_policy(write_format)
        self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

        # (2) Check for the presence of certain actions that we know will be there
        self.assertEqual(
            write_format["Statement"][0]["Action"],
            ["ssm:deletepar*", "ssm:getpar*", "ssm:la*", "ssm:putp*", "ssm:un*"],
            "extra actions are returned",
        )
        self.assertEqual(
            write_format["Statement"][0]["Resource"],
            cfg["read"],
            "Wrong resources were returned",
        )

    def test_minimize_rw_different(self):
        cfg = {
            "mode": "crud",
            "name": "",
            "read": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter2",
            ],
            "write": [
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter10",
            ],
        }
        sid_grp = SidGroup()
        write_format = sid_grp.process_template(cfg, minimize=1)
        """
        Expected result:

        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "SsmReadParameter",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:getpar*"
                    ],
                    "Resource": [
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter2"
                    ]
                },
                {
                    "Sid": "SsmWriteParameter",
                    "Effect": "Allow",
                    "Action": [
                        "ssm:deletepar*",
                        "ssm:la*",
                        "ssm:putp*"
                    ],
                    "Resource": [
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter",
                        "arn:aws:ssm:us-east-1:123456789012:parameter/myparameter10"
                    ]
                }
            ]
        }
        """
        self.assertEqual(
            len(write_format["Statement"]),
            2,
            "More than two statements returned, expected 2",
        )
        self.assertEqual(
            write_format["Statement"][0]["Action"],
            ["ssm:getpar*"],
            "extra actions are returned",
        )
        self.assertEqual(
            write_format["Statement"][0]["Resource"],
            cfg["read"],
            "Wrong resources were returned",
        )
        self.assertEqual(
            write_format["Statement"][1]["Action"],
            ["ssm:deletepar*", "ssm:la*", "ssm:putp*", "ssm:un*"],
            "extra actions are returned",
        )
        self.assertEqual(
            write_format["Statement"][1]["Resource"],
            cfg["write"],
            "Wrong resources were returned",
        )

    def test_minimize_arn_case_bucket(self):
        cfg = {
            "mode": "crud",
            "read": ["arn:aws:s3:::bucket_name"],
            "write": ["arn:aws:s3:::bucket_name"],
        }
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)

        # print(json.dumps(results, indent=4))
        # To future-proof this unit test...
        # (1) check that there is only one SID so it was combined during minimization
        sid_names = get_sid_names_from_policy(results)
        # For S3, it does require iam:PassRole lol because of s3:PutReplicationConfiguration requiring it as an action
        self.assertEqual(len(sid_names), 2, "More than 2 statements returned, expected 2")

    def test_minimize_arn_case_1(self):
        """minimization test with ARN types from test_does_arn_match_case_1"""
        cfg = {
            "mode": "crud",
            "read": ["arn:aws:codecommit:us-east-1:123456789012:MyDemoRepo"],
            "write": ["arn:aws:codecommit:us-east-1:123456789012:MyDemoRepo"],
        }
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)
        sid_names = get_sid_names_from_policy(results)
        self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

    def test_minimize_arn_case_3(self):
        """minimization test with ARN types from test_does_arn_match_case_3"""
        cfg = {
            "mode": "crud",
            "read": [
                "arn:aws:kinesis:us-east-1:account-id:firehose/myfirehose/consumer/someconsumer:${ConsumerCreationTimpstamp}"
            ],
            "write": [
                "arn:aws:kinesis:us-east-1:account-id:firehose/myfirehose/consumer/someconsumer:${ConsumerCreationTimpstamp}"
            ],
        }
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)
        sid_names = get_sid_names_from_policy(results)
        self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

    def test_minimize_arn_case_4(self):
        """minimization test with ARN types from test_does_arn_match_case_4"""
        cfg = {
            "mode": "crud",
            "read": ["arn:aws:batch:region:account-id:job-definition/job-name:revision"],
            "write": ["arn:aws:batch:region:account-id:job-definition/job-name:revision"],
        }
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)
        sid_names = get_sid_names_from_policy(results)
        self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

    def test_minimize_arn_case_5(self):
        """minimization test with ARN types from test_does_arn_match_case_5"""
        cfg = {
            "mode": "crud",
            "read": ["arn:aws:states:region:account-id:stateMachine:stateMachineName"],
            "write": ["arn:aws:states:region:account-id:stateMachine:stateMachineName"],
        }
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)
        sid_names = get_sid_names_from_policy(results)
        self.assertEqual(len(sid_names), 2, "More than two statements returned, expected 2")

    def test_minimize_arn_case_6(self):
        """minimization test with ARN types from test_does_arn_match_case_6"""
        cfg = {
            "mode": "crud",
            "read": ["arn:aws:states:region:account-id:execution:stateMachineName:executionName"],
            "write": ["arn:aws:states:region:account-id:execution:stateMachineName:executionName"],
        }
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)
        sid_names = get_sid_names_from_policy(results)
        self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

    # TODO: Greengrass now has two documentation pages. Need to fix that. Perhaps look to #226 for guidance? idk.
    # def test_minimize_arn_case_7(self):
    #     """minimization test with ARN types from test_does_arn_match_case_7"""
    #     cfg = {
    #         "mode": "crud",
    #         "read": ["arn:aws:greengrass:${Region}:${Account}:/greengrass/definition/devices/1234567/versions/1"],
    #         "write": ["arn:aws:greengrass:${Region}:${Account}:/greengrass/definition/devices/1234567/versions/1"]
    #     }
    #     sid_group = SidGroup()
    #     results = sid_group.process_template(cfg, minimize=0)
    #     sid_names = get_sid_names_from_policy(results)
    #     self.assertEqual(len(sid_names), 1, "More than one statement returned, expected 1")

    def test_GH_364_remove_duplicates_from_minimize(self):
        cfg = {
            "mode": "crud",
            "read": ["arn:aws:rds:*:*:*:test-*"],
            "write": ["arn:aws:rds:*:*:*:test-*"],
            "list": ["arn:aws:rds:*:*:*:test-*"],
            "tagging": ["arn:aws:rds:*:*:*:test-*"],
            "permissions-management": ["arn:aws:rds:*:*:*:test-*"],
        }
        # Bug introduced in https://github.com/salesforce/policy_sentry/pull/252
        sid_group = SidGroup()
        results = sid_group.process_template(cfg, minimize=0)
        print(json.dumps(results, indent=4))
        actions = results["Statement"][0]["Action"]
        actions_set = set(actions)
        self.assertEqual(
            len(actions_set),
            len(actions),
            "There should be no duplicate strings in the actions.",
        )



from policy_sentry.util.actions import (
    get_service_from_action,
    get_action_name_from_action,
    get_full_action_name,
)


def test_get_service_from_action():
    # given
    action = "ec2:DescribeInstance"

    # when
    service = get_service_from_action(action=action)

    # then
    assert service == "ec2"


def test_get_action_name_from_action():
    # given
    action = "ec2:DescribeInstance"

    # when
    service = get_action_name_from_action(action=action)

    # then
    assert service == "describeinstance"


def test_get_full_action_name():
    # given
    service = "ec2"
    action_name = "DescribeInstance"

    # when
    action = get_full_action_name(service=service, action_name=action_name)

    # then
    assert action == "ec2:DescribeInstance"



"""
Create YML Template files for the write-policy command.
Users don't have to remember exactly how to phrase the yaml files, so this command creates it for them.
"""

from __future__ import annotations

import logging
import sys
from pathlib import Path

import click

from policy_sentry import set_stream_logger
from policy_sentry.writing.template import create_actions_template, create_crud_template

logger = logging.getLogger(__name__)


@click.command(
    context_settings={"max_content_width": 160},
    short_help="Create write-policy YML template files",
)
@click.option(
    "--output-file",
    "-o",
    type=str,
    required=True,
    help="Relative path to output file where we want to store policy_sentry YML files.",
)
@click.option(
    "--template-type",
    "-t",
    type=click.Choice(["actions", "crud"], case_sensitive=False),
    required=True,
    help="Type of write_policy template to create - actions or CRUD. Case insensitive.",
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def create_template(output_file: str | Path, template_type: str, verbose: str) -> None:
    """
    Writes YML file templates for use in the write-policy
    command, so users can fill out the fields
    without needing to look up the required format.
    """
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)

    if template_type == "actions":
        template = create_actions_template()
    elif template_type == "crud":
        template = create_crud_template()
    else:
        print(f"Unknown template type: {template_type}")
        sys.exit()

    filename = Path(output_file).resolve()
    with filename.open("a", encoding="utf-8") as file_obj:
        file_obj.write(template)

    print(f"write-policy template file written to: {filename}")



import unittest
import json
from policy_sentry.querying.all import get_all_actions
from policy_sentry.querying.actions import get_actions_for_service
from policy_sentry.querying.arns import get_arn_types_for_service


class IAMDefinitionQA(unittest.TestCase):
    def setUp(self) -> None:
        all_actions = get_all_actions()
        self.all_actions = list(all_actions)
        self.all_actions.sort()

    def test_services_with_multiple_pages_apigateway(self):
        """Ensure that apigateway v1 and apigateway v2 actions are both present in the ses namespace"""
        # API Gateway Management V1: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonapigatewaymanagement.html
        self.assertTrue("apigateway:AddCertificateToDomain" in self.all_actions)
        self.assertTrue("apigateway:RemoveCertificateFromDomain" in self.all_actions)
        self.assertTrue("apigateway:SetWebACL" in self.all_actions)
        # API Gateway Management V2: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonapigatewaymanagement.html
        # API Gateway V2 doesn't have any unique actions in but it does have some unique resource types. Let's make sure those resource types are in the IAM Definition.
        # Resource types unique to API Gateway V2:
        resource_types = get_arn_types_for_service("apigateway")
        resource_types = list(resource_types.keys())
        self.assertTrue("AccessLogSettings" in resource_types)
        # Resource types unique to API Gateway V1:
        self.assertTrue("RestApi" in resource_types)

    def test_services_with_multiple_pages_aws_marketplace(self):
        """Ensure that aws-marketplace actions from all the different aws-marketplace SAR pages are present in the IAM definition."""
        # Overlap: AWS Marketplace, Marketplace Catalog, and AWS Marketplace Entitlement service, AWS Marketplace Image Building Service, AWS Marketplace Metering Service, AWS Marketplace Private Marketplace, and AWS Marketplace Procurement Systems
        # AWS Marketplace: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplace.html
        self.assertTrue("aws-marketplace:AcceptAgreementApprovalRequest" in self.all_actions)
        # AWS Marketplace Catalog: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplacecatalog.html
        self.assertTrue("aws-marketplace:CancelChangeSet" in self.all_actions)
        # AWS Marketplace Entitlement Service: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplaceentitlementservice.html
        self.assertTrue("aws-marketplace:GetEntitlements" in self.all_actions)
        # AWS Marketplace Image Building Service: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplaceimagebuildingservice.html
        self.assertTrue("aws-marketplace:DescribeBuilds" in self.all_actions)
        # AWS Marketplace Metering Service: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplacemeteringservice.html
        self.assertTrue("aws-marketplace:BatchMeterUsage" in self.all_actions)
        # AWS Marketplace Private Marketplace: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplaceprivatemarketplace.html
        self.assertTrue("aws-marketplace:AssociateProductsWithPrivateMarketplace" in self.all_actions)
        # AWS Marketplace Procurement Systems: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsmarketplaceprocurementsystemsintegration.html
        self.assertTrue("aws-marketplace:DescribeProcurementSystemConfiguration" in self.all_actions)

        results = get_actions_for_service("aws-marketplace")
        actions = [
            "aws-marketplace:AcceptAgreementApprovalRequest",
            "aws-marketplace:BatchMeterUsage",
            "aws-marketplace:CancelAgreementRequest",
            "aws-marketplace:CancelChangeSet",
            "aws-marketplace:DescribeAgreement",
            "aws-marketplace:DescribeBuilds",
            "aws-marketplace:DescribeChangeSet",
            "aws-marketplace:DescribeEntity",
            "aws-marketplace:DescribeProcurementSystemConfiguration",
            "aws-marketplace:GetAgreementApprovalRequest",
            "aws-marketplace:GetAgreementRequest",
            "aws-marketplace:GetAgreementTerms",
            "aws-marketplace:GetEntitlements",
            "aws-marketplace:ListAgreementApprovalRequests",
            "aws-marketplace:ListAgreementRequests",
            "aws-marketplace:ListBuilds",
            "aws-marketplace:ListChangeSets",
            "aws-marketplace:ListEntities",
            "aws-marketplace:MeterUsage",
            "aws-marketplace:PutProcurementSystemConfiguration",
            "aws-marketplace:RegisterUsage",
            "aws-marketplace:RejectAgreementApprovalRequest",
            "aws-marketplace:ResolveCustomer",
            "aws-marketplace:SearchAgreements",
            "aws-marketplace:StartBuild",
            "aws-marketplace:StartChangeSet",
            "aws-marketplace:Subscribe",
            "aws-marketplace:Unsubscribe",
            "aws-marketplace:UpdateAgreementApprovalRequest",
            "aws-marketplace:ViewSubscriptions",
        ]
        for action in actions:
            self.assertIn(action, results)

    def test_services_with_multiple_pages_greengrass(self):
        """Ensure that greengrass v1 and greengrass v2 actions are both present in the greengrass namespace"""
        # Greengrass V1: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiotgreengrass.html
        self.assertTrue("greengrass:CreateResourceDefinition" in self.all_actions)
        # Greengrass V2: https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiotgreengrassv2.html
        self.assertTrue("greengrass:CreateComponentVersion" in self.all_actions)
        results = get_actions_for_service("greengrass")
        actions = [
            "greengrass:AssociateRoleToGroup",
            "greengrass:CreateConnectorDefinition",
            "greengrass:CreateConnectorDefinitionVersion",
            "greengrass:CreateCoreDefinition",
            "greengrass:CreateCoreDefinitionVersion",
            "greengrass:CreateDeviceDefinition",
            "greengrass:CreateDeviceDefinitionVersion",
            "greengrass:CreateFunctionDefinition",
            "greengrass:CreateFunctionDefinitionVersion",
            "greengrass:CreateGroup",
            "greengrass:CreateGroupCertificateAuthority",
            "greengrass:CreateGroupVersion",
            "greengrass:CreateLoggerDefinition",
            "greengrass:CreateLoggerDefinitionVersion",
            "greengrass:CreateResourceDefinition",
            "greengrass:CreateResourceDefinitionVersion",
            "greengrass:CreateSoftwareUpdateJob",
            "greengrass:CreateSubscriptionDefinition",
            "greengrass:CreateSubscriptionDefinitionVersion",
            "greengrass:DeleteConnectorDefinition",
            "greengrass:DeleteCoreDefinition",
            "greengrass:DeleteDeviceDefinition",
            "greengrass:DeleteFunctionDefinition",
            "greengrass:DeleteGroup",
            "greengrass:DeleteLoggerDefinition",
            "greengrass:DeleteResourceDefinition",
            "greengrass:DeleteSubscriptionDefinition",
            "greengrass:DisassociateRoleFromGroup",
            "greengrass:Discover",
            "greengrass:GetAssociatedRole",
            "greengrass:GetBulkDeploymentStatus",
            "greengrass:GetConnectorDefinition",
            "greengrass:GetConnectorDefinitionVersion",
            "greengrass:GetCoreDefinition",
            "greengrass:GetCoreDefinitionVersion",
            "greengrass:GetDeploymentStatus",
            "greengrass:GetDeviceDefinition",
            "greengrass:GetDeviceDefinitionVersion",
            "greengrass:GetFunctionDefinition",
            "greengrass:GetFunctionDefinitionVersion",
            "greengrass:GetGroup",
            "greengrass:GetGroupCertificateAuthority",
            "greengrass:GetGroupCertificateConfiguration",
            "greengrass:GetGroupVersion",
            "greengrass:GetLoggerDefinition",
            "greengrass:GetLoggerDefinitionVersion",
            "greengrass:GetResourceDefinition",
            "greengrass:GetResourceDefinitionVersion",
            "greengrass:GetSubscriptionDefinition",
            "greengrass:GetSubscriptionDefinitionVersion",
            "greengrass:GetThingRuntimeConfiguration",
            "greengrass:ListBulkDeploymentDetailedReports",
            "greengrass:ListBulkDeployments",
            "greengrass:ListConnectorDefinitionVersions",
            "greengrass:ListConnectorDefinitions",
            "greengrass:ListCoreDefinitionVersions",
            "greengrass:ListCoreDefinitions",
            "greengrass:ListDeviceDefinitionVersions",
            "greengrass:ListDeviceDefinitions",
            "greengrass:ListFunctionDefinitionVersions",
            "greengrass:ListFunctionDefinitions",
            "greengrass:ListGroupCertificateAuthorities",
            "greengrass:ListGroupVersions",
            "greengrass:ListGroups",
            "greengrass:ListLoggerDefinitionVersions",
            "greengrass:ListLoggerDefinitions",
            "greengrass:ListResourceDefinitionVersions",
            "greengrass:ListResourceDefinitions",
            "greengrass:ListSubscriptionDefinitionVersions",
            "greengrass:ListSubscriptionDefinitions",
            "greengrass:ResetDeployments",
            "greengrass:StartBulkDeployment",
            "greengrass:StopBulkDeployment",
            "greengrass:UpdateConnectorDefinition",
            "greengrass:UpdateCoreDefinition",
            "greengrass:UpdateDeviceDefinition",
            "greengrass:UpdateFunctionDefinition",
            "greengrass:UpdateGroup",
            "greengrass:UpdateGroupCertificateConfiguration",
            "greengrass:UpdateLoggerDefinition",
            "greengrass:UpdateResourceDefinition",
            "greengrass:UpdateSubscriptionDefinition",
            "greengrass:UpdateThingRuntimeConfiguration",
        ]
        for action in actions:
            self.assertIn(action, results)

    def test_services_with_multiple_pages_elb(self):
        """Ensure that elb v1 and elb v2 actions are both present in the elasticloadbalancing namespace"""
        results = get_actions_for_service("elasticloadbalancing")
        actions = [
            "elasticloadbalancing:ApplySecurityGroupsToLoadBalancer",
            "elasticloadbalancing:AttachLoadBalancerToSubnets",
            "elasticloadbalancing:ConfigureHealthCheck",
            "elasticloadbalancing:CreateAppCookieStickinessPolicy",
            "elasticloadbalancing:CreateLBCookieStickinessPolicy",
            "elasticloadbalancing:CreateLoadBalancerListeners",
            "elasticloadbalancing:CreateLoadBalancerPolicy",
            "elasticloadbalancing:DeleteLoadBalancerListeners",
            "elasticloadbalancing:DeleteLoadBalancerPolicy",
            "elasticloadbalancing:DeregisterInstancesFromLoadBalancer",
            "elasticloadbalancing:DescribeInstanceHealth",
            "elasticloadbalancing:DescribeLoadBalancerPolicies",
            "elasticloadbalancing:DescribeLoadBalancerPolicyTypes",
            "elasticloadbalancing:DetachLoadBalancerFromSubnets",
            "elasticloadbalancing:DisableAvailabilityZonesForLoadBalancer",
            "elasticloadbalancing:EnableAvailabilityZonesForLoadBalancer",
            "elasticloadbalancing:RegisterInstancesWithLoadBalancer",
            "elasticloadbalancing:SetLoadBalancerListenerSSLCertificate",
            "elasticloadbalancing:SetLoadBalancerPoliciesForBackendServer",
            "elasticloadbalancing:SetLoadBalancerPoliciesOfListener",
        ]
        for action in actions:
            self.assertIn(action, results)

    def test_services_with_multiple_pages_lex(self):
        """Ensure that lex v1 and lex v2 actions are both present in the lex namespace"""
        # Lex V1: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonlex.html
        self.assertTrue("lex:DeleteUtterances" in self.all_actions)
        # Lex V2: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonlexv2.html
        self.assertTrue("lex:ListBotLocales" in self.all_actions)
        results = get_actions_for_service("lex")
        actions = [
            "lex:CreateIntentVersion",
            "lex:CreateSlotTypeVersion",
            "lex:DeleteBotChannelAssociation",
            "lex:DeleteIntentVersion",
            "lex:DeleteSlotTypeVersion",
            "lex:GetBot",
            "lex:GetBotAlias",
            "lex:GetBotAliases",
            "lex:GetBotChannelAssociation",
            "lex:GetBotChannelAssociations",
            "lex:GetBotVersions",
            "lex:GetBots",
            "lex:GetBuiltinIntent",
            "lex:GetBuiltinIntents",
            "lex:GetBuiltinSlotTypes",
            "lex:GetExport",
            "lex:GetImport",
            "lex:GetIntent",
            "lex:GetIntentVersions",
            "lex:GetIntents",
            "lex:GetMigration",
            "lex:GetMigrations",
            "lex:GetSlotType",
            "lex:GetSlotTypeVersions",
            "lex:GetSlotTypes",
            "lex:GetUtterancesView",
            "lex:PostContent",
            "lex:PostText",
            "lex:PutBot",
            "lex:PutBotAlias",
            "lex:PutIntent",
            "lex:PutSlotType",
            "lex:StartMigration",
        ]
        for action in actions:
            self.assertIn(action, results)

    def test_services_with_multiple_pages_kinesis_analytics(self):
        """Ensure that Kinesis Analytics V1 actions are both present in the ses namespace"""
        # Kinesis Analytics V1
        results = get_actions_for_service("kinesisanalytics")
        actions = [
            "kinesisanalytics:GetApplicationState",  # Only in v1, not v2
            "kinesisanalytics:ListApplications",  # In both
        ]
        for action in actions:
            self.assertIn(action, results)

    def test_services_with_multiple_pages_ses(self):
        """Ensure that ses v1 and ses v2 actions are both present in the ses namespace"""
        # SES V1: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonses.html
        self.assertTrue("ses:PutIdentityPolicy" in self.all_actions)
        # SES V2: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonsimpleemailservicev2.html
        self.assertTrue("ses:ListImportJobs" in self.all_actions)

        results = get_actions_for_service("ses")
        actions = [
            "ses:CloneReceiptRuleSet",
            "ses:CreateConfigurationSetTrackingOptions",
            "ses:CreateReceiptFilter",
            "ses:CreateReceiptRule",
            "ses:CreateReceiptRuleSet",
            "ses:CreateTemplate",
            "ses:DeleteConfigurationSetTrackingOptions",
            "ses:DeleteIdentity",
            "ses:DeleteIdentityPolicy",
            "ses:DeleteReceiptFilter",
            "ses:DeleteReceiptRule",
            "ses:DeleteReceiptRuleSet",
            "ses:DeleteTemplate",
            "ses:DeleteVerifiedEmailAddress",
            "ses:DescribeActiveReceiptRuleSet",
            "ses:DescribeConfigurationSet",
            "ses:DescribeReceiptRule",
            "ses:DescribeReceiptRuleSet",
            "ses:GetAccountSendingEnabled",
            "ses:GetIdentityDkimAttributes",
            "ses:GetIdentityMailFromDomainAttributes",
            "ses:GetIdentityNotificationAttributes",
            "ses:GetIdentityPolicies",
            "ses:GetIdentityVerificationAttributes",
            "ses:GetSendQuota",
            "ses:GetSendStatistics",
            "ses:GetTemplate",
            "ses:ListIdentities",
            "ses:ListIdentityPolicies",
            "ses:ListReceiptFilters",
            "ses:ListReceiptRuleSets",
            "ses:ListTemplates",
            "ses:ListVerifiedEmailAddresses",
            "ses:PutIdentityPolicy",
            "ses:ReorderReceiptRuleSet",
            "ses:SendBounce",
            "ses:SendBulkTemplatedEmail",
            "ses:SendRawEmail",
            "ses:SendTemplatedEmail",
            "ses:SetActiveReceiptRuleSet",
            "ses:SetIdentityDkimEnabled",
            "ses:SetIdentityFeedbackForwardingEnabled",
            "ses:SetIdentityHeadersInNotificationsEnabled",
            "ses:SetIdentityMailFromDomain",
            "ses:SetIdentityNotificationTopic",
            "ses:SetReceiptRulePosition",
            "ses:TestRenderTemplate",
            "ses:UpdateAccountSendingEnabled",
            "ses:UpdateConfigurationSetReputationMetricsEnabled",
            "ses:UpdateConfigurationSetSendingEnabled",
            "ses:UpdateConfigurationSetTrackingOptions",
            "ses:UpdateReceiptRule",
            "ses:UpdateTemplate",
            "ses:VerifyDomainDkim",
            "ses:VerifyDomainIdentity",
            "ses:VerifyEmailAddress",
            "ses:VerifyEmailIdentity",
        ]
        for action in actions:
            self.assertIn(action, results)


class IAMDefinitionQAForServicesWithChangedHTMLFiles(unittest.TestCase):
    def setUp(self) -> None:
        all_actions = get_all_actions()
        self.all_actions = list(all_actions)
        self.all_actions.sort()

    def test_other_iam_data_fixes_in_GH_393(self):
        """Other missing actions from GH #393"""
        # Cassandra: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazonkeyspacesforapachecassandra.html
        results = get_actions_for_service("cassandra")
        self.assertTrue("cassandra:Restore" in results)
        # Comprehend Medical: https://docs.aws.amazon.com/service-authorization/latest/reference/list_amazoncomprehendmedical.html
        results = get_actions_for_service("comprehendmedical")
        # print(results)
        actions = [
            "comprehendmedical:DescribeEntitiesDetectionV2Job",
            "comprehendmedical:DescribeICD10CMInferenceJob",
            "comprehendmedical:DescribePHIDetectionJob",
            "comprehendmedical:DescribeRxNormInferenceJob",
            # "comprehendmedical:DescribeSNOMEDCTInferenceJob",  # Not in SAR
            "comprehendmedical:DetectEntitiesV2",
            "comprehendmedical:InferICD10CM",
            "comprehendmedical:InferRxNorm",
            # "comprehendmedical:InferSNOMEDCT",  # Not in SAR
            "comprehendmedical:ListEntitiesDetectionV2Jobs",
            "comprehendmedical:ListICD10CMInferenceJobs",
            "comprehendmedical:ListPHIDetectionJobs",
            "comprehendmedical:ListRxNormInferenceJobs",
            # "comprehendmedical:ListSNOMEDCTInferenceJobs",  # Not in SAR
            "comprehendmedical:StartEntitiesDetectionV2Job",
            "comprehendmedical:StartICD10CMInferenceJob",
            "comprehendmedical:StartPHIDetectionJob",
            "comprehendmedical:StartRxNormInferenceJob",
            "comprehendmedical:StopEntitiesDetectionV2Job",
            "comprehendmedical:StopICD10CMInferenceJob",
        ]
        for action in actions:
            self.assertIn(action, results)
        # Compute Optimizer
        results = get_actions_for_service("compute-optimizer")
        actions = [
            "compute-optimizer:DeleteRecommendationPreferences",
            "compute-optimizer:ExportEBSVolumeRecommendations",
            "compute-optimizer:ExportLambdaFunctionRecommendations",
            "compute-optimizer:GetEffectiveRecommendationPreferences",
            "compute-optimizer:GetEnrollmentStatusesForOrganization",
            "compute-optimizer:GetLambdaFunctionRecommendations",
            "compute-optimizer:GetRecommendationPreferences",
            "compute-optimizer:PutRecommendationPreferences",
        ]
        for action in actions:
            self.assertIn(action, results)
        # DataSync
        results = get_actions_for_service("datasync")
        actions = [
            "datasync:UpdateLocationNfs",
            "datasync:UpdateLocationObjectStorage",
            "datasync:UpdateLocationSmb",
            "datasync:UpdateTaskExecution",
        ]
        for action in actions:
            self.assertIn(action, results)

        # Account Management
        results = get_actions_for_service("account")
        actions = [
            "account:DeleteAlternateContact",
            "account:GetAlternateContact",
            "account:PutAlternateContact",
        ]
        for action in actions:
            self.assertIn(action, results)

        # AWS IAM Access Analyzer
        results = get_actions_for_service("access-analyzer")
        actions = [
            "access-analyzer:CancelPolicyGeneration",
            "access-analyzer:CreateAccessPreview",
            "access-analyzer:GetAccessPreview",
            "access-analyzer:GetGeneratedPolicy",
            "access-analyzer:ListAccessPreviewFindings",
            "access-analyzer:ListAccessPreviews",
            "access-analyzer:ListPolicyGenerations",
            "access-analyzer:StartPolicyGeneration",
            "access-analyzer:ValidatePolicy",
        ]
        for action in actions:
            self.assertIn(action, results)
        # Elemental Activations
        results = get_actions_for_service("elemental-activations")
        actions = [
            "elemental-activations:CompleteAccountRegistration",
            "elemental-activations:StartFileUpload",
        ]
        for action in actions:
            self.assertIn(action, results)
        # OpenSearch
        results = get_actions_for_service("es")
        actions = [
            "es:DescribeDomainChangeProgress",
        ]
        for action in actions:
            self.assertIn(action, results)
        # Location
        results = get_actions_for_service("geo")
        actions = [
            "geo:CalculateRouteMatrix",
        ]
        for action in actions:
            self.assertIn(action, results)

        # Amazon Managed Grafana
        results = get_actions_for_service("grafana")
        actions = [
            "grafana:DescribeWorkspaceAuthentication",
            "grafana:UpdateWorkspaceAuthentication",
        ]
        for action in actions:
            self.assertIn(action, results)

        # EC2 Image Builder
        results = get_actions_for_service("imagebuilder")
        actions = [
            "imagebuilder:ImportVmImage",
        ]
        for action in actions:
            self.assertIn(action, results)
        # Timestream
        results = get_actions_for_service("timestream")
        actions = [
            "timestream:CreateScheduledQuery",
            "timestream:DeleteScheduledQuery",
            "timestream:DescribeScheduledQuery",
            "timestream:ExecuteScheduledQuery",
            "timestream:ListScheduledQueries",
            "timestream:UpdateScheduledQuery",
        ]
        for action in actions:
            self.assertIn(action, results)

        # AWS Transfer Family
        results = get_actions_for_service("transfer")
        actions = [
            "transfer:CreateAccess",
            "transfer:CreateWorkflow",
            "transfer:DeleteAccess",
            "transfer:DeleteWorkflow",
            "transfer:DescribeAccess",
            "transfer:DescribeExecution",
            "transfer:DescribeWorkflow",
            "transfer:ListAccesses",
            "transfer:ListExecutions",
            "transfer:ListWorkflows",
            "transfer:SendWorkflowStepState",
            "transfer:UpdateAccess",
        ]
        for action in actions:
            self.assertIn(action, results)

    def test_kafka_action_names_overlap_issue(self):
        """Ensure that kafka actions are not overwritten in the IAM definition"""
        # Kafka actions used to be in two pages but are now one. This verifies the current state.
        # results = get_actions_for_service("kafka")
        # print(results)
        actions = [
            "kafka:BatchAssociateScramSecret",
            "kafka:BatchDisassociateScramSecret",
            "kafka:CreateClusterV2",
            "kafka:DeleteConfiguration",
            "kafka:DescribeClusterV2",
            "kafka:ListClustersV2",
            "kafka:ListConfigurationRevisions",
            "kafka:ListKafkaVersions",
            "kafka:ListScramSecrets",
            "kafka:RebootBroker",
            "kafka:UpdateBrokerType",
            "kafka:UpdateConfiguration",
            "kafka:UpdateConnectivity",
            "kafka:UpdateSecurity",
        ]

        for action in actions:
            self.assertIn(action, self.all_actions)



import os
import json
import unittest
from click.testing import CliRunner
from policy_sentry.command.write_policy import write_policy
from policy_sentry.util.policy_files import get_sid_names_from_policy

test_file_directory = os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, "examples", "yml")


class PolicySentryClickUnitTests(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def test_write_policy_basic_with_click(self):
        """command.write_policy: using crud.yml should return exit code 0"""
        template_file = os.path.join(test_file_directory, "crud.yml")
        result = self.runner.invoke(write_policy, ["--input-file", template_file])
        self.assertTrue(result.exit_code == 0)

    def test_click_crud_case_1(self):
        """write_policy: using 1-ssm-read.yml"""
        template_file = os.path.join(test_file_directory, "crud-cases", "1-ssm-read.yml")
        result = self.runner.invoke(write_policy, ["--input-file", template_file])
        result_json = json.loads(result.output)
        self.assertListEqual(get_sid_names_from_policy(result_json), ["SsmReadParameter"])
        self.assertTrue(result.exit_code == 0)

    def test_click_crud_case_2(self):
        """write_policy: using 2-skip-resource-constraints.yml"""
        template_file = os.path.join(test_file_directory, "crud-cases", "2-skip-resource-constraints.yml")
        result = self.runner.invoke(write_policy, ["--input-file", template_file])
        result_json = json.loads(result.output)
        self.assertListEqual(get_sid_names_from_policy(result_json), ["SkipResourceConstraints"])
        self.assertTrue(result.exit_code == 0)

    def test_click_crud_case_3(self):
        """write_policy: using 3-wildcard-only-single-actions.yml"""
        template_file = os.path.join(test_file_directory, "crud-cases", "3-wildcard-only-single-actions.yml")
        result = self.runner.invoke(write_policy, ["--input-file", template_file])
        result_json = json.loads(result.output)
        self.assertListEqual(get_sid_names_from_policy(result_json), ["MultMultNone"])
        self.assertTrue(result.exit_code == 0)

    def test_click_crud_case_4(self):
        """write_policy: using 4-wildcard-only-bulk-selection.yml"""
        template_file = os.path.join(test_file_directory, "crud-cases", "4-wildcard-only-bulk-selection.yml")
        result = self.runner.invoke(write_policy, ["--input-file", template_file])
        result_json = json.loads(result.output)
        self.assertListEqual(get_sid_names_from_policy(result_json), ["MultMultNone"])
        self.assertTrue(result.exit_code == 0)



"""
sid_group indicates that this is a collection of policy-related data organized by their SIDs
"""

from __future__ import annotations

import logging
import re
from typing import Any

from policy_sentry.analysis.expand import determine_actions_to_expand
from policy_sentry.querying.actions import (
    get_action_data,
    get_actions_at_access_level_that_support_wildcard_arns_only,
    get_actions_that_support_wildcard_arns_only,
    get_actions_with_arn_type_and_access_level,
    get_dependent_actions,
)
from policy_sentry.querying.all import get_all_actions
from policy_sentry.querying.arns import get_resource_type_name_with_raw_arn
from policy_sentry.shared.constants import POLICY_LANGUAGE_VERSION
from policy_sentry.util.actions import get_lowercase_action_list
from policy_sentry.util.arns import does_arn_match, get_service_from_arn, parse_arn
from policy_sentry.util.text import capitalize_first_character, strip_special_characters
from policy_sentry.writing.minimize import minimize_statement_actions
from policy_sentry.writing.validate import check_actions_schema, check_crud_schema

SANITIZE_NAME_PATTERN = re.compile(r"[^A-Za-z0-9]+")

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class SidGroup:
    """
    This class is critical to the creation of least privilege policies.
    It uses the SIDs as namespaces. The namespaces follow this format:
        `{Servicename}{Accesslevel}{Resourcetypename}`

    So, a resulting statement's SID might look like 'S3ListBucket'

    If a condition key is supplied (like `s3:RequestJob`), the SID string will be significantly longer.

    It will resemble this format:
            `{Servicename}{Accesslevel}{Resourcetypename}{Conditionkeystring}{Conditiontypestring}{Conditionkeyvalue}`

    For example: EC2 write actions on the security-group resource, using the following condition map:
    ```
        "Condition": {
            "StringEquals": {"ec2:ResourceTag/Owner": "${aws:username}"}
        }
    ```

    The resulting SID would be:
        `Ec2WriteSecuritygroupResourcetagownerStringequalsAwsusername`

    Or, for actions that support wildcard ARNs only, an example could be:
        `Ec2WriteMultResourcetagownerStringequalsAwsusername`
    """

    def __init__(self) -> None:
        # Dict instead of list
        # sids instead of ARN
        self.sids: dict[str, dict[str, Any]] = {}
        self.universal_conditions: dict[str, Any] = {}
        self.skip_resource_constraints: list[str] = []
        self.exclude_actions: list[str] = []
        self.wildcard_only_single_actions: list[str] = []
        # When a user requests all wildcard-only actions available under a service at a specific access level
        self.wildcard_only_service_read: list[str] = []
        self.wildcard_only_service_write: list[str] = []
        self.wildcard_only_service_list: list[str] = []
        self.wildcard_only_service_tagging: list[str] = []
        self.wildcard_only_service_permissions_management: list[str] = []

    def get_sid_group(self) -> dict[str, dict[str, Any]]:
        """
        Get the whole SID group as JSON
        """
        return self.sids

    def get_sid(self, sid: str) -> dict[str, Any]:
        """Get a single group by the SID identifier"""
        if self.sids[sid]:
            return self.sids[sid]

        raise Exception(f"No SID with the value of {sid}")

    def list_sids(self) -> list[str]:
        """
        Get a list of all of the SIDs by their identifiers

        Returns:
            List: A list of SIDs in the SID group
        """
        return list(self.sids.keys())

    def add_exclude_actions(self, exclude_actions: list[str]) -> None:
        """To exclude actions from the output"""
        if exclude_actions:
            expanded_actions = determine_actions_to_expand(exclude_actions)
            self.exclude_actions = [x.lower() for x in expanded_actions]
        else:
            self.exclude_actions = []

    def add_skip_resource_constraints(self, skip_resource_constraints_actions: str | list[str]) -> None:
        """
        To override resource constraint requirements - i.e., instead of restricting `s3:PutObject` to a path and
        allowing `s3:PutObject` to `*` resources, put `s3:GetObject` here.
        """
        if isinstance(skip_resource_constraints_actions, list):
            self.skip_resource_constraints.extend(skip_resource_constraints_actions)
        elif isinstance(skip_resource_constraints_actions, str):
            self.skip_resource_constraints.append(skip_resource_constraints_actions)
        else:
            raise Exception("Please provide 'skip_resource_constraints' as a list of IAM actions.")

    def add_sts_actions(self, sts_actions: dict[str, list[str]]) -> None:
        """
        To add STS actions to the output from special YAML section
        """
        if sts_actions:
            # Hard coded for this special case
            service_prefix = "sts"
            access_level = "Write"

            for action, arns in sts_actions.items():
                clean_action = action.replace("-", "")  # Convention to follow adding dashes instead of CamelCase
                service_action_data = get_action_data(service_prefix, clean_action)

                # Schema validation takes care of this, but just in case. No data returned for the action
                if not service_action_data:
                    raise Exception(f"Could not find service action data for {service_prefix} - {clean_action}")

                for row in service_action_data[service_prefix]:
                    for arn in arns:
                        if not arn:  # skip the - '' situation
                            continue
                        if row["access_level"] == access_level and does_arn_match(arn, row["resource_arn_format"]):
                            raw_arn_format = row["resource_arn_format"]

                            # Each action will get its own namespace sts:AssumeRole -> AssumeRole
                            # -1 index is a neat trick if the colon ever goes away we won't get an index error.
                            sid_namespace = row["action"].split(":")[-1]

                            temp_sid_dict = {
                                "arn": [arn],
                                "service": service_prefix,
                                "access_level": access_level,
                                "arn_format": raw_arn_format,
                                "actions": [row["action"]],
                                "conditions": [],  # TODO: Add conditions
                            }

                            # Using a custom namespace and not gathering actions so no need to find
                            # dependent actions either, though we could do it here

                            if sid_namespace in self.sids:
                                # If the ARN already exists there, skip it.
                                if arn not in self.sids[sid_namespace]["arn"]:
                                    self.sids[sid_namespace]["arn"].append(arn)
                            else:
                                self.sids[sid_namespace] = temp_sid_dict

    def add_requested_service_wide(self, service_prefixes: list[str], access_level: str) -> None:
        """
        When a user requests all wildcard-only actions available under a service at a specific access level

        Arguments:
            service_prefixes: A list of service prefixes
            access_level: The requested access level
        """
        if access_level == "Read":
            self.wildcard_only_service_read = service_prefixes
        elif access_level == "Write":
            self.wildcard_only_service_write = service_prefixes
        elif access_level == "List":
            self.wildcard_only_service_list = service_prefixes
        elif access_level == "Tagging":
            self.wildcard_only_service_tagging = service_prefixes
        elif access_level == "Permissions management":
            self.wildcard_only_service_permissions_management = service_prefixes

    def process_wildcard_only_actions(self) -> None:
        """
        After (1) the list of wildcard-only single actions have been added and (2) the list of wildcard-only service-wide actions have been added, process them and store them under the proper SID.
        """
        provided_wildcard_actions = (
            self.wildcard_only_single_actions
            + get_wildcard_only_actions_matching_services_and_access_level(self.wildcard_only_service_read, "Read")
            + get_wildcard_only_actions_matching_services_and_access_level(self.wildcard_only_service_list, "List")
            + get_wildcard_only_actions_matching_services_and_access_level(self.wildcard_only_service_write, "Write")
            + get_wildcard_only_actions_matching_services_and_access_level(
                self.wildcard_only_service_tagging, "Tagging"
            )
            + get_wildcard_only_actions_matching_services_and_access_level(
                self.wildcard_only_service_permissions_management,
                "Permissions management",
            )
        )
        self.add_wildcard_only_actions(provided_wildcard_actions)

    def get_rendered_policy(self, minimize: int | None = None) -> dict[str, Any]:
        """
        Get the JSON rendered policy

        Arguments:
            minimize: Reduce the character count of policies without creating overlap with other action names
        Returns:
            Dictionary: The IAM Policy JSON
        """
        statements: list[dict[str, Any]] = []
        # Only set the actions to lowercase if minimize is provided
        all_actions = get_all_actions(lowercase=True)

        # render the policy
        sids_to_be_changed = []
        for sid, group in self.sids.items():
            temp_actions = group["actions"]
            if not temp_actions:
                logger.debug(f"No actions for sid {sid}")
                continue
            actions = []
            if self.exclude_actions:
                for temp_action in temp_actions:
                    if temp_action.lower() in self.exclude_actions:
                        logger.debug(f"\tExcluded action: {temp_action}")
                    else:
                        if temp_action not in actions:
                            actions.append(temp_action)
            else:
                actions = temp_actions
            # Check if SID is empty of actions. Continue if yes.
            if not actions:
                continue
            match_found = False
            if minimize is not None and isinstance(minimize, int):
                logger.debug("Minimizing statements...")
                actions = minimize_statement_actions(actions, all_actions, minchars=minimize)
                # searching in the existing statements
                # further minimizing the output
                for stmt in statements:
                    if stmt["Resource"] == group["arn"]:
                        stmt["Action"].extend(x for x in actions if x not in stmt["Action"])
                        match_found = True
                        sids_to_be_changed.append(stmt["Sid"])
                        break
                actions = list(set(actions))  # remove duplicates
            actions.sort()
            logger.debug(f"Adding statement with SID {sid}")
            logger.debug(f"{sid} SID has the actions: {actions}")
            logger.debug(f"{sid} SID has the resources: {group['arn']}")
            if not match_found:
                statements.append(
                    {
                        "Sid": sid,
                        "Effect": "Allow",
                        "Action": actions,
                        "Resource": group["arn"],
                    }
                )

        if sids_to_be_changed:
            for stmt in statements:
                if stmt["Sid"] in sids_to_be_changed:
                    arn_details = parse_arn(stmt["Resource"][0])
                    resource_path = arn_details.get("resource_path")
                    resource_sid_segment = strip_special_characters(f"{arn_details['resource']}{resource_path}")
                    stmt["Sid"] = create_policy_sid_namespace(arn_details["service"], "Mult", resource_sid_segment)
                    # If we have combined the statements, minimize it again
                    if minimize is not None and isinstance(minimize, int):
                        actions = minimize_statement_actions(stmt["Action"], all_actions, minchars=minimize)
                        actions.sort()
                        stmt["Action"] = actions
        return {"Version": POLICY_LANGUAGE_VERSION, "Statement": statements}

    # pylint: disable=unused-argument
    def add_by_arn_and_access_level(
        self,
        arn_list: list[str],
        access_level: str,
        _conditions_block: dict[str, Any] | None = None,
    ) -> None:
        """
        This adds the user-supplied ARN(s), service prefixes, access levels, and condition keys (if applicable) given
        by the user. It derives the list of IAM actions based on the user's requested ARNs and access levels.

        Arguments:
            arn_list: Just a list of resource ARNs.
            access_level: "Read", "List", "Tagging", "Write", or "Permissions management"
            conditions_block: Optionally, a condition block with one or more conditions
        """
        for arn in arn_list:
            service_prefix = get_service_from_arn(arn)
            if "$" in service_prefix:
                logger.debug(f"Not supported service {service_prefix} found in ARN {arn}")
                continue

            service_action_data = get_action_data(service_prefix, "*")
            for service_prefix, action_data in service_action_data.items():
                for row in action_data:
                    raw_arn_format = row["resource_arn_format"]
                    if row["access_level"] == access_level and does_arn_match(arn, raw_arn_format):
                        resource_type_name = get_resource_type_name_with_raw_arn(raw_arn_format)
                        if resource_type_name is None:
                            continue

                        sid_namespace = create_policy_sid_namespace(service_prefix, access_level, resource_type_name)
                        actions = get_actions_with_arn_type_and_access_level(
                            service_prefix, resource_type_name, access_level
                        )
                        # List comprehension to get all dependent actions that are not in the supplied actions.
                        dependent_actions = [
                            action for action in get_dependent_actions(actions) if action not in actions
                        ]
                        for dep_action in dependent_actions:
                            self.add_action_without_resource_constraint(dep_action)

                        if sid_namespace in self.sids:
                            # If the ARN already exists there, skip it.
                            if arn not in self.sids[sid_namespace]["arn"]:
                                self.sids[sid_namespace]["arn"].append(arn)
                        # If it did not exist before at all, create it.
                        else:
                            temp_sid_dict = {
                                "arn": [arn],
                                "service": service_prefix,
                                "access_level": access_level,
                                "arn_format": raw_arn_format,
                                "actions": actions,
                                "conditions": [],  # TODO: Add conditions
                            }
                            self.sids[sid_namespace] = temp_sid_dict

    def add_action_without_resource_constraint(
        self, action: str, sid_namespace: str = "MultMultNone"
    ) -> dict[str, Any]:
        """
        This handles the cases where certain actions do not handle resource constraints - either by AWS, or for
        flexibility when adding dependent actions.

        Arguments:
            action: The single action to add to the SID namespace. For instance, s3:ListAllMyBuckets
            sid_namespace: MultMultNone by default. Other valid option is "SkipResourceConstraints"
        """
        if sid_namespace == "SkipResourceConstraints":
            temp_sid_dict = {
                "arn": ["*"],
                "service": "Skip",
                "access_level": "ResourceConstraints",
                "arn_format": "*",
                "actions": [action],
            }
        elif sid_namespace == "MultMultNone":
            temp_sid_dict = {
                "arn": ["*"],
                "service": "Mult",
                "access_level": "Mult",
                "arn_format": "*",
                "actions": [action],
            }
        else:
            raise Exception("Please specify the sid_namespace as either 'SkipResourceConstraints' or 'MultMultNone'.")
        if isinstance(action, str):
            if sid_namespace in self.sids:
                if action not in self.sids[sid_namespace]["actions"]:
                    self.sids[sid_namespace]["actions"].append(action)
            else:
                self.sids[sid_namespace] = temp_sid_dict
        else:
            raise Exception("Please provide the action as a string, not a list.")
        return self.sids

    def add_by_list_of_actions(self, supplied_actions: list[str]) -> dict[str, Any]:
        """
        Takes a list of actions, queries the database for corresponding arns, adds them to the object.

        Arguments:
            supplied_actions: A list of supplied actions
        """
        dependent_actions = [
            action for action in get_dependent_actions(supplied_actions) if action not in supplied_actions
        ]
        logger.debug("Adding by list of actions")
        logger.debug(f"Supplied actions: {supplied_actions}")
        logger.debug(f"Dependent actions: {dependent_actions}")
        arns_matching_supplied_actions = []

        # arns_matching_supplied_actions is a list of dicts.
        # It must do this rather than dictionaries because there will be duplicate
        #     values by nature of how the entries in the IAM database are structured.
        # I'll provide the example values here to improve readability.

        for action in supplied_actions:
            service_name, action_name = action.split(":")
            action_data = get_action_data(service_name, action_name)
            for row in action_data[service_name]:
                if row["resource_arn_format"] not in arns_matching_supplied_actions:
                    arns_matching_supplied_actions.append(
                        {
                            "resource_arn_format": row["resource_arn_format"],
                            "access_level": row["access_level"],
                            "action": row["action"],
                        }
                    )

        # Identify the actions that do not support resource constraints
        # If that's the case, add it to the wildcard namespace. Otherwise, don't add it.

        actions_without_resource_constraints = []
        for item in arns_matching_supplied_actions:
            if item["resource_arn_format"] != "*":
                self.add_by_arn_and_access_level([item["resource_arn_format"]], item["access_level"])
            else:
                actions_without_resource_constraints.append(item["action"])

        # If there are any dependent actions, we need to add them without resource constraints.
        # Otherwise, we get into issues where the amount of extra SIDs will balloon.
        # Also, the user has no way of knowing what those dependent actions are beforehand.
        # TODO: This is, in fact, a great opportunity to introduce conditions. But we aren't there yet.
        for dep_action in dependent_actions:
            self.add_action_without_resource_constraint(dep_action)
        # Now, because add_by_arn_and_access_level() adds all actions under an access level, we have to
        # remove all actions that do not match the supplied_actions. This is done in-place.
        logger.debug("Purging actions that do not match the requested actions and dependent actions")
        logger.debug(f"Supplied actions: {supplied_actions}")
        logger.debug(f"Dependent actions: {dependent_actions}")
        self.remove_actions_not_matching_these(supplied_actions + dependent_actions)
        for action in actions_without_resource_constraints:
            logger.debug(f"Deliberately adding the action {action} without resource constraints")
            self.add_action_without_resource_constraint(action)
        logger.debug(
            "Removing actions that are in the wildcard arn (Resources = '*') as well as other statements that have "
            "resource constraints "
        )
        self.remove_actions_duplicated_in_wildcard_arn()
        logger.debug("Getting the rendered policy")
        return self.get_rendered_policy()

    def process_template(self, cfg: dict[str, Any], minimize: int | None = None) -> dict[str, Any]:
        """
        Process the Policy Sentry template as a dict. This auto-detects whether or not the file is in CRUD mode or
        Actions mode.

        Arguments:
            cfg: The loaded YAML as a dict. Must follow Policy Sentry dictated format.
            minimize: Minimize the resulting statement with *safe* usage of wildcards to reduce policy length. Set this to the character length you want - for example, 0, or 4. Defaults to none.
        Returns:
            Dictionary: The rendered IAM JSON Policy
        """
        cfg_mode = cfg.get("mode")
        if cfg_mode == "crud":
            logger.debug("CRUD mode selected")
            check_crud_schema(cfg)
            # EXCLUDE ACTIONS
            cfg_exclude = cfg.get("exclude-actions")
            if cfg_exclude and cfg_exclude[0]:
                self.add_exclude_actions(cfg_exclude)
            # WILDCARD ONLY SECTION
            cfg_wildcard = cfg.get("wildcard-only")
            if cfg_wildcard:
                provided_wildcard_actions = cfg_wildcard.get("single-actions")
                if provided_wildcard_actions and provided_wildcard_actions[0] != "":
                    logger.debug(f"Requested wildcard-only actions: {provided_wildcard_actions}")
                    self.wildcard_only_single_actions = provided_wildcard_actions

                service_read = cfg_wildcard.get("service-read")
                if service_read and service_read[0]:
                    logger.debug(f"Requested wildcard-only actions: {service_read}")
                    self.wildcard_only_service_read = service_read

                service_write = cfg_wildcard.get("service-write")
                if service_write and service_write[0]:
                    logger.debug(f"Requested wildcard-only actions: {service_write}")
                    self.wildcard_only_service_write = service_write

                service_list = cfg_wildcard.get("service-list")
                if service_list and service_list[0]:
                    logger.debug(f"Requested wildcard-only actions: {service_list}")
                    self.wildcard_only_service_list = service_list

                service_tagging = cfg_wildcard.get("service-tagging")
                if service_tagging and service_tagging[0]:
                    logger.debug(f"Requested wildcard-only actions: {service_tagging}")
                    self.wildcard_only_service_tagging = service_tagging

                service_permissions_management = cfg_wildcard.get("service-permissions-management")
                if service_permissions_management and service_permissions_management[0]:
                    logger.debug(f"Requested wildcard-only actions: {service_permissions_management}")
                    self.wildcard_only_service_permissions_management = service_permissions_management

            # Process the wildcard-only section
            self.process_wildcard_only_actions()

            # Standard access levels
            cfg_read = cfg.get("read")
            if cfg_read and cfg_read[0]:
                logger.debug(f"Requested access to arns: {cfg_read}")
                self.add_by_arn_and_access_level(cfg_read, "Read")
            cfg_write = cfg.get("write")
            if cfg_write and cfg_write[0]:
                logger.debug(f"Requested access to arns: {cfg_write}")
                self.add_by_arn_and_access_level(cfg_write, "Write")
            cfg_list = cfg.get("list")
            if cfg_list and cfg_list[0]:
                logger.debug(f"Requested access to arns: {cfg_list}")
                self.add_by_arn_and_access_level(cfg_list, "List")
            tagging = cfg.get("tagging")
            if tagging and tagging[0]:
                logger.debug(f"Requested access to arns: {tagging}")
                self.add_by_arn_and_access_level(tagging, "Tagging")
            cfg_mgmt = cfg.get("permissions-management")
            if cfg_mgmt and cfg_mgmt[0]:
                logger.debug(f"Requested access to arns: {cfg_mgmt}")
                self.add_by_arn_and_access_level(cfg_mgmt, "Permissions management")

            # SKIP RESOURCE CONSTRAINTS
            cfg_skip = cfg.get("skip-resource-constraints")
            if cfg_skip and cfg_skip[0]:
                logger.debug(f"Requested override: the actions {cfg_skip} will skip resource constraints.")
                self.add_skip_resource_constraints(cfg_skip)
                for skip_resource_constraints_action in self.skip_resource_constraints:
                    self.add_action_without_resource_constraint(
                        skip_resource_constraints_action, "SkipResourceConstraints"
                    )

            # STS Section
            cfg_sts = cfg.get("sts")
            if cfg_sts:
                logger.debug("STS section detected. Building assume role policy statement")
                self.add_sts_actions(cfg_sts)

        elif cfg_mode == "actions":
            check_actions_schema(cfg)
            if "actions" in cfg:
                cfg_actions = cfg["actions"]
                if cfg_actions is not None and cfg_actions[0] != "":
                    self.add_by_list_of_actions(cfg_actions)

        return self.get_rendered_policy(minimize)

    def add_wildcard_only_actions(self, provided_wildcard_actions: list[str]) -> None:
        """
        Given a list of IAM actions, add individual IAM Actions that do not support resource constraints to the MultMultNone SID

        Arguments:
            provided_wildcard_actions: list actions provided by the user.
        """
        if isinstance(provided_wildcard_actions, list):
            verified_wildcard_actions = remove_actions_that_are_not_wildcard_arn_only(provided_wildcard_actions)
            if verified_wildcard_actions:
                logger.debug(f"Attempting to add the following actions to the policy: {verified_wildcard_actions}")
                self.add_by_list_of_actions(verified_wildcard_actions)
                logger.debug(f"Added the following wildcard-only actions to the policy: {verified_wildcard_actions}")

    def add_wildcard_only_actions_matching_services_and_access_level(
        self, services: list[str], access_level: str
    ) -> None:
        """
        Arguments:
            services: A list of AWS services
            access_level: An access level as it is written in the database, such as 'Read', 'Write', 'List', 'Permissions management', or 'Tagging'
        """
        wildcard_only_actions_to_add = []
        for service in services:
            actions = get_actions_at_access_level_that_support_wildcard_arns_only(service, access_level)
            wildcard_only_actions_to_add.extend(actions)
        self.add_wildcard_only_actions(wildcard_only_actions_to_add)

    def remove_actions_not_matching_these(self, actions_to_keep: list[str]) -> None:
        """
        Arguments:
            actions_to_keep: A list of actions to leave in the policy. All actions not in this list are removed.
        """
        actions_to_keep = get_lowercase_action_list(actions_to_keep)
        actions_deleted = []
        for group in self.sids.values():
            placeholder_actions_list = []
            for action in group["actions"]:
                # if the action is not in the list of selected actions, don't copy it to the placeholder list
                action_lower = action.lower()
                if action_lower in actions_to_keep:
                    placeholder_actions_list.append(action)
                else:
                    logger.debug(f"{action_lower} not found in list of actions to keep: {actions_to_keep}")
                    actions_deleted.append(action)
            # only include the updated actions
            group["actions"] = placeholder_actions_list
        # Highlight the actions that you remove
        logger.debug(f"Actions deleted: {actions_deleted}")
        # Now that we've removed a bunch of actions, if there are SID groups without any actions,
        # remove them so we don't get SIDs with empty action lists
        self.remove_sids_with_empty_action_lists()

    def remove_sids_with_empty_action_lists(self) -> None:
        """
        Now that we've removed a bunch of actions, if there are SID groups without any actions, remove them so we don't get SIDs with empty action lists
        """
        sid_namespaces_to_delete = [
            sid
            for sid, group in self.sids.items()
            if not group["actions"]  # If the size is zero, add it to the indexes_to_delete list.
        ]
        # Loop through sid_namespaces_to_delete in reverse order (so we delete index
        # 10 before index 8, for example)
        for sid in reversed(sid_namespaces_to_delete):
            del self.sids[sid]

    def remove_actions_duplicated_in_wildcard_arn(self) -> None:
        """
        Removes actions from the object that are in a resource-specific ARN, as well as the `*` resource.
        For example, if `ssm:GetParameter` is restricted to a specific parameter path, as well as `*`, then we want to
        remove the `*` option to force least privilege.
        """
        actions_under_wildcard_resources = []
        actions_under_wildcard_resources_to_nuke = []

        # Build a temporary list. Contains actions in MultMultNone SID (where resources = "*")
        for group in self.sids.values():
            if group["arn_format"] == "*":
                actions_under_wildcard_resources.extend(group["actions"])

        # If the actions under the MultMultNone SID exist under other SIDs
        if actions_under_wildcard_resources:
            for group in self.sids.values():
                if "*" not in group["arn_format"]:
                    for action in actions_under_wildcard_resources:
                        if action in group["actions"] and action not in self.skip_resource_constraints:
                            # add it to a list of actions to nuke when they are under other SIDs
                            actions_under_wildcard_resources_to_nuke.append(action)  # noqa: PERF401

        # If there are actions that we need to remove from SIDs outside of MultMultNone SID
        if actions_under_wildcard_resources_to_nuke:
            for group in self.sids.values():
                if "*" in group["arn_format"]:
                    for action in actions_under_wildcard_resources_to_nuke:
                        if action in group["actions"]:
                            group["actions"].remove(action)


def remove_actions_that_are_not_wildcard_arn_only(actions_list: list[str]) -> list[str]:
    """
    Given a list of actions, remove the ones that CAN be restricted to ARNs, leaving only the ones that cannot.

    Arguments:
        actions_list: A list of actions
    Returns:
        List: An updated list of actions
    """
    # remove duplicates, if there are any
    actions_set = set(actions_list)
    actions_list_placeholder = []

    for action in actions_set:
        try:
            service_name, _ = action.split(":")
        except ValueError as v_e:
            # We will skip the action because this likely means that the wildcard action provided is not valid.
            logger.debug(v_e)
            logger.debug("The value provided in wildcard-only section is not formatted properly.")
            continue

        action_lower = action.lower()
        rows = get_actions_that_support_wildcard_arns_only(service_name)
        for row in rows:
            if row.lower() == action_lower:
                actions_list_placeholder.append(action)  # noqa: PERF401
    return actions_list_placeholder


def get_wildcard_only_actions_matching_services_and_access_level(services: list[str], access_level: str) -> list[str]:
    """
    Get a list of wildcard-only actions matching the services and access level

    Arguments:
        services: A list of AWS services
        access_level: An access level as it is written in the database, such as 'Read', 'Write', 'List', 'Permissions management', or 'Tagging'
    Returns:
        List: A list of wildcard-only actions matching the services and access level
    """
    wildcard_only_actions_to_add = []
    for service in services:
        actions = get_actions_at_access_level_that_support_wildcard_arns_only(service, access_level)
        wildcard_only_actions_to_add.extend(actions)
    return wildcard_only_actions_to_add


def create_policy_sid_namespace(
    service: str,
    access_level: str,
    resource_type_name: str,
    condition_block: dict[str, Any] | None = None,
) -> str:
    """
    Simply generates the SID name. The SID groups ARN types that share an access level.

    For example, S3 objects vs. SSM Parameter have different ARN types - as do S3 objects vs S3 buckets. That's how we
    choose to group them.

    Arguments:
        service: `ssm`
        access_level: `Read`
        resource_type_name: `parameter`
        condition_block: `{"condition_key_string": "ec2:ResourceTag/purpose", "condition_type_string": "StringEquals", "condition_value": "test"}`

    Returns:
        String: A string like `SsmReadParameter`
    """
    # Sanitize the resource_type_name; otherwise we hit some list conversion
    # errors
    resource_type_name = re.sub(SANITIZE_NAME_PATTERN, "", resource_type_name)
    # Also remove the space from the Access level, if applicable. This only
    # applies for "Permissions management"
    access_level = re.sub(SANITIZE_NAME_PATTERN, "", access_level)
    sid_namespace_prefix = (
        capitalize_first_character(strip_special_characters(service))
        + capitalize_first_character(access_level)
        + capitalize_first_character(resource_type_name)
    )

    if condition_block:
        condition_key_namespace = re.sub(SANITIZE_NAME_PATTERN, "", condition_block["condition_key_string"])
        condition_type_namespace = condition_block["condition_type_string"]
        condition_value_namespace = re.sub(SANITIZE_NAME_PATTERN, "", condition_block["condition_value"])
        sid_namespace_condition_suffix = (
            f"{capitalize_first_character(condition_key_namespace)}"
            f"{capitalize_first_character(condition_type_namespace)}"
            f"{capitalize_first_character(condition_value_namespace)}"
        )
        sid_namespace = sid_namespace_prefix + sid_namespace_condition_suffix
    else:
        sid_namespace = sid_namespace_prefix
    return sid_namespace



"""
Just a common storage space for storing some constants.
"""

import logging
import os
from pathlib import Path

logger = logging.getLogger()

# General Folders
HOME = Path.home()
CONFIG_DIRECTORY = HOME / ".policy_sentry"

# HTML Docs
BUNDLED_HTML_DIRECTORY_PATH = Path(__file__).parent / "data/docs"
BUNDLED_DATA_DIRECTORY = Path(__file__).parent / "data"

LOCAL_HTML_DIRECTORY_PATH = CONFIG_DIRECTORY / "data/docs"

BASE_DOCUMENTATION_URL = "https://docs.aws.amazon.com/service-authorization/latest/reference/reference_policies_actions-resources-contextkeys.html"
# BASE_DOCUMENTATION_URL = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_actions-resources-contextkeys.html"

# Data json file
# On initialization, load the IAM data
BUNDLED_DATASTORE_FILE_PATH = Path(__file__).parent / "data/iam-definition.json"
LOCAL_DATASTORE_FILE_PATH = CONFIG_DIRECTORY / "iam-definition.json"
# Check for the existence of the local datastore first.
if LOCAL_DATASTORE_FILE_PATH.exists():
    # If it exists, leverage that datastore instead of the one bundled with the python package
    logger.info(
        f"Leveraging the local IAM definition at the path: {LOCAL_DATASTORE_FILE_PATH} "
        f"To leverage the bundled definition instead, remove the folder $HOME/.policy_sentry/"
    )
    DATASTORE_FILE_PATH = LOCAL_DATASTORE_FILE_PATH
else:
    # Otherwise, leverage the datastore inside the python package
    logger.debug("Leveraging the bundled IAM Definition.")
    DATASTORE_FILE_PATH = BUNDLED_DATASTORE_FILE_PATH

# Overrides
if "CUSTOM_ACCESS_OVERRIDES_FILE" in os.environ:
    CUSTOM_ACCESS_OVERRIDES_FILE = os.environ["CUSTOM_ACCESS_OVERRIDES_FILE"]
    BUNDLED_ACCESS_OVERRIDES_FILE = Path(__file__).parent / CUSTOM_ACCESS_OVERRIDES_FILE

else:
    BUNDLED_ACCESS_OVERRIDES_FILE = Path(__file__).parent / "data/access-level-overrides.yml"

LOCAL_ACCESS_OVERRIDES_FILE = CONFIG_DIRECTORY / "access-level-overrides.yml"

# Policy constants
# https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_version.html
POLICY_LANGUAGE_VERSION = "2012-10-17"

# IAM datastore schema versions
POLICY_SENTRY_SCHEMA_VERSION_NAME = "policy_sentry_schema_version"
POLICY_SENTRY_SCHEMA_VERSION_V2 = "v2"
POLICY_SENTRY_SCHEMA_VERSION_LATEST = POLICY_SENTRY_SCHEMA_VERSION_V2



"""
Methods that execute specific queries against the SQLite database for the SERIVCES table.
This supports the policy_sentry query functionality
"""

from __future__ import annotations

from policy_sentry.shared.iam_data import iam_definition


def get_services_data() -> list[dict[str, str]]:
    return [
        {
            "prefix": service_prefix,
            "service_name": data["service_name"],
        }
        for service_prefix, data in iam_definition.items()
        if isinstance(data, dict)
    ]



"""
Validation for the Policy Sentry YML Templates.
"""

from __future__ import annotations

import logging
from typing import Any

from schema import And, Optional, Regex, Schema, SchemaError, Use

logger = logging.getLogger(__name__)


def check(conf_schema: Schema, conf: dict[str, Any]) -> bool:
    """
    Validates a user-supplied JSON vs a defined schema.

    Arguments:
        conf_schema: The Schema object that defines the required structure.
        conf: The user-supplied schema to validate against the required structure.
    Returns:
        Boolean: The decision about whether the JSON meets expected Schema requirements
    """
    try:
        conf_schema.validate(conf)
        return True
    except SchemaError as schema_error:
        try:
            # workarounds for Schema's logging approach
            print(schema_error.autos[0])
            detailed_error_message = schema_error.autos[2]
            print(detailed_error_message.split(" in {'")[0])
            # for error in schema_error.autos:
        except:  # noqa: E722
            logger.critical(schema_error)
        return False


CRUD_SCHEMA = Schema(
    {
        "mode": "crud",
        Optional("name"): And(Use(str)),
        Optional("read"): [str],
        Optional("write"): [str],
        Optional("list"): [str],
        Optional("permissions-management"): [str],
        Optional("tagging"): [str],
        Optional("wildcard-only"): {
            Optional("single-actions"): [str],
            Optional("service-read"): [str],
            Optional("service-write"): [str],
            Optional("service-list"): [str],
            Optional("service-tagging"): [str],
            Optional("service-permissions-management"): [str],
        },
        Optional("skip-resource-constraints"): [str],
        Optional("exclude-actions"): [str],
        Optional("sts"): {And(Use(str), Regex(r"^assume-role(-with-)*(saml|web-identity)*$")): [str]},
    }
)

ACTIONS_SCHEMA = Schema(
    {
        "mode": "actions",
        Optional("name"): And(Use(str)),
        "actions": And([str]),
    }
)


def check_actions_schema(cfg: dict[str, Any]) -> bool:
    """
    Determines whether the user-provided config matches the required schema for Actions mode
    """
    result = check(ACTIONS_SCHEMA, cfg)
    if result is True:
        return result

    raise Exception(
        "The provided template does not match the required schema for ACTIONS mode. "
        "Please use the create-template command to generate a valid YML template that "
        "Policy Sentry will accept."
    )


def check_crud_schema(cfg: dict[str, Any]) -> bool:
    """
    Determines whether the user-provided config matches the required schema for CRUD mode
    """
    result = check(CRUD_SCHEMA, cfg)
    if result is True:
        return result

    raise Exception(
        "The provided template does not match the required schema for CRUD mode. "
        "Please use the create-template command to generate a valid YML template that "
        "Policy Sentry will accept."
    )


def validate_condition_block(condition_block: dict[str, Any]) -> bool:
    """
    Validates the format of the condition block that should be supplied in the template.

    Arguments:
        condition_block: {"condition_key_string": "ec2:ResourceTag/purpose", "condition_type_string": "StringEquals", "condition_value": "test"}
    Returns:
        Boolean: The decision
    """

    # TODO: Validate that the values are legit somehow
    condition_block_schema = Schema(
        {
            "condition_key_string": And(Use(str)),
            "condition_type_string": And(Use(str)),
            "condition_value": And(Use(str)),
        }
    )
    try:
        condition_block_schema.validate(condition_block)
        # TODO: Try to validate whether or not the condition keys are legit
        return True
    except SchemaError as s_e:
        logger.warning(s_e)
        return False



import unittest
import json
import os
from policy_sentry.command.write_policy import write_policy_with_template
from policy_sentry.writing.sid_group import SidGroup
from policy_sentry.writing.template import (
    get_crud_template_dict,
    get_actions_template_dict,
)
from policy_sentry.util.arns import ARN

desired_crud_policy = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "MultMultNone",
            "Effect": "Allow",
            "Action": ["cloudhsm:DescribeClusters", "kms:CreateCustomKeyStore"],
            "Resource": ["*"],
        },
        {
            "Sid": "SecretsmanagerReadSecret",
            "Effect": "Allow",
            "Action": [
                "secretsmanager:DescribeSecret",
                "secretsmanager:GetResourcePolicy",
                "secretsmanager:GetSecretValue",
                "secretsmanager:ListSecretVersionIds",
            ],
            "Resource": ["arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret"],
        },
        {
            "Sid": "SecretsmanagerWriteSecret",
            "Effect": "Allow",
            "Action": [
                "secretsmanager:CancelRotateSecret",
                "secretsmanager:CreateSecret",
                "secretsmanager:DeleteSecret",
                "secretsmanager:PutSecretValue",
                "secretsmanager:RestoreSecret",
                "secretsmanager:RotateSecret",
                "secretsmanager:UpdateSecret",
                "secretsmanager:UpdateSecretVersionStage",
            ],
            "Resource": ["arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret"],
        },
        {
            "Sid": "S3ListObject",
            "Effect": "Allow",
            "Action": ["s3:ListMultipartUploadParts"],
            "Resource": ["arn:aws:s3:::example-org-sbx-vmimport/stuff"],
        },
        {
            "Sid": "SsmTaggingParameter",
            "Effect": "Allow",
            "Action": ["ssm:AddTagsToResource", "ssm:RemoveTagsFromResource"],
            "Resource": ["arn:aws:ssm:us-east-1:123456789012:parameter/test"],
        },
        {
            "Sid": "KmsPermissionsmanagementKey",
            "Effect": "Allow",
            "Action": [
                "kms:CreateGrant",
                "kms:PutKeyPolicy",
                "kms:RetireGrant",
                "kms:RevokeGrant",
            ],
            "Resource": ["arn:aws:kms:us-east-1:123456789012:key/123456"],
        },
        {
            "Sid": "AssumeRole",
            "Effect": "Allow",
            "Action": ["sts:AssumeRole"],
            "Resource": ["arn:aws:iam::123456789012:role/demo"],
        },
    ],
}

desired_actions_policy = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "KmsPermissionsmanagementKey",
            "Effect": "Allow",
            "Action": ["kms:CreateGrant"],
            "Resource": ["arn:${Partition}:kms:${Region}:${Account}:key/${KeyId}"],
        },
        {
            "Sid": "MultMultNone",
            "Effect": "Allow",
            "Action": ["cloudhsm:DescribeClusters", "kms:CreateCustomKeyStore"],
            "Resource": ["*"],
        },
        {
            "Sid": "Ec2WriteSecuritygroup",
            "Effect": "Allow",
            "Action": [
                "ec2:AuthorizeSecurityGroupEgress",
                "ec2:AuthorizeSecurityGroupIngress",
            ],
            "Resource": ["arn:${Partition}:ec2:${Region}:${Account}:security-group/${SecurityGroupId}"],
        },
    ],
}


class WritePolicyWithLibraryOnly(unittest.TestCase):
    def test_write_actions_policy_with_library_only(self):
        """test_write_actions_policy_with_library_only: Write an actions mode policy without using the command line at all (library only)"""
        actions_template = get_actions_template_dict()
        # print(actions_template)
        actions_to_add = [
            "kms:CreateGrant",
            "kms:CreateCustomKeyStore",
            "ec2:AuthorizeSecurityGroupEgress",
            "ec2:AuthorizeSecurityGroupIngress",
        ]
        actions_template["mode"] = "actions"
        actions_template["actions"].extend(actions_to_add)
        # Modify it
        sid_group = SidGroup()
        minimize = None
        policy = sid_group.process_template(actions_template, minimize=minimize)
        self.maxDiff = None
        print(json.dumps(policy, indent=4))
        expected_statement_ids = [
            "KmsPermissionsmanagementKey",
            "MultMultNone",
            "Ec2WriteSecuritygroup",
            "Ec2WriteSecuritygrouprule",
        ]
        # self.assertDictEqual(policy, desired_actions_policy)
        for statement in policy.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)

    def test_write_crud_policy_with_library_only(self):
        """test_write_crud_policy_with_library_only: Write a policy in CRUD mode without using the command line at all (library only)"""
        another_crud_template = get_crud_template_dict()
        wildcard_actions_to_add = [
            "kms:CreateCustomKeyStore",
            # "cloudhsm:describeclusters",
        ]
        another_crud_template["mode"] = "crud"
        another_crud_template["read"].append("arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret")
        another_crud_template["write"].append("arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret")
        another_crud_template["list"].append("arn:aws:s3:::example-org-sbx-vmimport/stuff")
        another_crud_template["permissions-management"].append("arn:aws:kms:us-east-1:123456789012:key/123456")
        another_crud_template["tagging"].append("arn:aws:ssm:us-east-1:123456789012:parameter/test")
        another_crud_template["wildcard-only"]["single-actions"].extend(wildcard_actions_to_add)
        another_crud_template["sts"]["assume-role"].append("arn:aws:iam::123456789012:role/demo")
        # Modify it
        sid_group = SidGroup()
        # minimize = None
        result = sid_group.process_template(another_crud_template)
        expected_statement_ids = [
            "MultMultNone",
            "SecretsmanagerReadSecret",
            "SecretsmanagerWriteSecret",
            "S3ListObject",
            "SsmTaggingParameter",
            "KmsPermissionsmanagementKey",
            "AssumeRole",
        ]
        self.maxDiff = None
        print(json.dumps(result, indent=4))
        for statement in result.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)

    def test_gh_211_write_with_empty_access_level_lists(self):
        crud_template = get_crud_template_dict()
        crud_template["read"].append("arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret")
        crud_template["write"].append("arn:aws:secretsmanager:us-east-1:123456789012:secret:mysecret")
        # crud_template['list'].append("arn:aws:s3:::mybucket/stuff")
        # by commenting out the line below, you should not get an IndexError
        # crud_template['permissions-management'].append("arn:aws:kms:us-east-1:123456789012:key/123456")
        # crud_template['tagging'].append("arn:aws:ssm:us-east-1:123456789012:parameter/test")
        wildcard_actions_to_add = [
            "kms:createcustomkeystore",
            "cloudhsm:describeclusters",
        ]
        crud_template["wildcard-only"]["single-actions"].extend(wildcard_actions_to_add)
        result = write_policy_with_template(crud_template)
        # print(json.dumps(result, indent=4))
        expected_statement_ids = [
            "MultMultNone",
            "SecretsmanagerReadSecret",
            "SecretsmanagerWriteSecret",
        ]
        for statement in result.get("Statement"):
            self.assertTrue(statement.get("Sid") in expected_statement_ids)

    def test_gh_204_multiple_resource_types_wildcards(self):
        """test_gh_204_multiple_resource_types_wildcards: Address github issue #204 not having unit tests"""
        crud_template = {
            "mode": "crud",
            "read": ["arn:aws:rds:us-east-1:123456789012:*:*"],
            "write": ["arn:aws:rds:us-east-1:123456789012:*:*"],
            "list": ["arn:aws:rds:us-east-1:123456789012:*:*"],
        }

        # Let's only check the read level ones, or that will get exhausting.
        expected_statement_ids = [
            "RdsReadDb",
            "RdsReadEs",
            "RdsReadOg",
            "RdsReadPg",
            "RdsReadProxy",
            "RdsReadRi",
            "RdsReadSecgrp",
            "RdsReadSnapshot",
            "RdsReadSubgrp",
            "RdsReadTargetgroup",
            "MultMultNone",
        ]
        # In the real world, we would want to minimize, but that would just result in two different Sids:
        #   RdsMult
        #   MultMultNone
        # And in this test, we are trying to verify them individually to make sure they are not excluded.
        # So we will skip --minimize just for this test.
        policy = write_policy_with_template(crud_template)
        statement_ids = []
        for statement in policy.get("Statement"):
            statement_ids.append(statement.get("Sid"))
        for expected_sid in expected_statement_ids:
            self.assertTrue(expected_sid in statement_ids)

    def test_gh_237_ssm_arns_with_paths(self):
        """test_gh_237_ssm_arns_with_paths: Test GitHub issue #204 with resource ARN paths"""
        crud_template = {
            "mode": "crud",
            "read": ["arn:aws:ssm:::parameter/dev/foo/bar*"],
        }
        # result = write_policy_with_template(crud_template)
        # print(json.dumps(result, indent=4))
        arn = ARN("arn:aws:ssm:::parameter/dev/foo/bar*")
        self.assertTrue(arn.same_resource_type("arn:aws:ssm:::parameter/dev"))



"""
Allow users to use specific pre-compiled queries against the action, arn, and condition tables from command line.
"""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Any

import click
import yaml

from policy_sentry import set_stream_logger
from policy_sentry.querying.actions import (
    get_action_data,
    get_actions_for_service,
    get_actions_matching_arn_type,
    get_actions_matching_condition_key,
    get_actions_that_support_wildcard_arns_only,
    get_actions_with_access_level,
    get_actions_with_arn_type_and_access_level,
)
from policy_sentry.querying.all import get_all_service_prefixes
from policy_sentry.querying.arns import (
    get_arn_type_details,
    get_arn_types_for_service,
    get_raw_arns_for_service,
)
from policy_sentry.querying.conditions import (
    get_condition_key_details,
    get_condition_keys_for_service,
)
from policy_sentry.querying.services import get_services_data
from policy_sentry.shared.constants import (
    DATASTORE_FILE_PATH,
    LOCAL_DATASTORE_FILE_PATH,
)
from policy_sentry.util.access_levels import transform_access_level_text

if TYPE_CHECKING:
    from collections.abc import Iterable

logger = logging.getLogger(__name__)
iam_definition_path = DATASTORE_FILE_PATH


def print_list(output: Iterable[Any], fmt: str = "json") -> None:
    """Common method on how to print a list, depending on whether the user requests JSON or YAML output"""
    print(yaml.dump(output)) if fmt == "yaml" else [print(item) for item in output]


def print_dict(output: list[Any] | dict[Any, Any], fmt: str = "json") -> None:
    """Common method on how to print a dict, depending on whether the user requests JSON, YAML or CSV output"""
    if fmt == "csv":
        if not output:
            return
        print(",".join(output[0].keys()))
        for entry in output:
            print(",".join(entry.values()))
    elif fmt == "json":
        print(json.dumps(output, indent=4))
    elif fmt == "yaml":
        print(yaml.dump(output))


@click.group()
def query() -> None:
    """Allow users to query the IAM tables from command line"""


@query.command(
    short_help="Query the action table based on access levels, conditions, or actions that only support wildcard "
    "resources."
)
@click.option("--service", "-s", type=str, required=True, help="Filter according to AWS service.")
@click.option(
    "--name",
    "-n",
    type=str,
    required=False,
    help='The name of IAM Action. For example, if the action is "iam:ListUsers", supply "ListUsers" here.',
)
@click.option(
    "--access-level",
    "-a",
    type=click.Choice(["read", "write", "list", "tagging", "permissions-management"]),
    required=False,
    help="Filter according to CRUD levels. Acceptable values are read, write, list, tagging, permissions-management",
)
@click.option(
    "--condition",
    "-c",
    type=str,
    required=False,
    help="Supply a condition key to show a list of all IAM actions that support the condition key.",
)
@click.option(
    "--resource-type",
    "-r",
    type=str,
    required=False,
    help="Supply a resource type to show a list of all IAM actions that support the resource type.",
)
@click.option(
    "--fmt",
    type=click.Choice(["yaml", "json"]),
    default="json",
    required=False,
    help='Format output as YAML or JSON. Defaults to "json"',
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def action_table(
    name: str,
    service: str,
    access_level: str,
    condition: str,
    resource_type: str,
    fmt: str,
    verbose: str | None,
) -> None:
    """Query the Action Table from the Policy Sentry database"""
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)

    query_action_table(name, service, access_level, condition, resource_type, fmt)


def query_action_table(
    name: str,
    service: str,
    access_level: str,
    condition: str,
    resource_type: str,
    fmt: str = "json",
) -> Iterable[str]:
    """Query the Action Table from the Policy Sentry database.
    Use this one when leveraging Policy Sentry as a library."""
    if LOCAL_DATASTORE_FILE_PATH.exists():
        logger.info(
            f"Using the Local IAM definition: {LOCAL_DATASTORE_FILE_PATH}. To leverage the bundled definition instead, remove the folder $HOME/.policy_sentry/"
        )
    else:
        # Otherwise, leverage the datastore inside the python package
        logger.debug("Leveraging the bundled IAM Definition.")

    if service == "catalog":
        # make sure nothing is returned when querying for `catalog`,
        # which has some special behaviour related to `servicecatalog`
        service = ""

    # Actions on all services
    if service == "all":
        all_services = get_all_service_prefixes()
        if access_level:
            level = transform_access_level_text(access_level)
            print(f"{access_level} actions across ALL services:\n")
            output = []
            for serv in all_services:
                result = get_actions_with_access_level(serv, level)
                output.extend(result)
            print_list(output=output, fmt=fmt)
        # Get a list of all services in the database
        elif resource_type == "*":
            print("ALL actions that do not support resource ARN constraints")
            output = get_actions_that_support_wildcard_arns_only(service)
            print_dict(output=output, fmt=fmt)
        else:
            print("All services in the database:\n")
            # it is a set here, which is ok
            output = all_services
            print_list(output=output, fmt=fmt)
    elif name is None and access_level and not resource_type:
        print(f"All IAM actions under the {service} service that have the access level {access_level}:")
        level = transform_access_level_text(access_level)
        output = get_actions_with_access_level(service, level)
        print_dict(output=output, fmt=fmt)
    elif name is None and access_level and resource_type:
        print(f"{service} {access_level.upper()} actions that have the resource type {resource_type.upper()}:")
        access_level = transform_access_level_text(access_level)
        output = get_actions_with_arn_type_and_access_level(service, resource_type, access_level)
        print_list(output=output, fmt=fmt)
    # Get a list of all IAM actions under the service that support the specified condition key.
    elif condition:
        print(f"IAM actions under {service} service that support the {condition} condition only:")
        output = get_actions_matching_condition_key(service, condition)
        print_list(output=output, fmt=fmt)
    # Get a list of IAM Actions under the service that only support resources = "*"
    # (i.e., you cannot restrict it according to ARN)
    elif resource_type:
        print(f"IAM actions under {service} service that have the resource type {resource_type}:")
        output = get_actions_matching_arn_type(service, resource_type)
        print_dict(output=output, fmt=fmt)
    elif name and access_level is None:
        output = get_action_data(service, name)
        print_dict(output=output, fmt=fmt)
    else:
        # Get a list of all IAM Actions available to the service
        output = get_actions_for_service(service)
        print(f"ALL {service} actions:")
        print_list(output=output, fmt=fmt)
    return output


@query.command(
    short_help="Query the ARN table to show RAW ARNs, like `aws:s3:::bucket/object`. "
    "Use --list-arn-types ARN types, like `object`."
)
@click.option("--service", "-s", type=str, required=True, help="Filter according to AWS service.")
@click.option(
    "--name",
    "-n",
    type=str,
    required=False,
    help="The short name of the resource ARN type. For example, `bucket` under service `s3`.",
)
@click.option(
    "--list-arn-types",
    "-l",
    is_flag=True,
    required=False,
    help="Show the short names of ARN Types. If empty, this will show RAW ARNs only.",
)
@click.option(
    "--fmt",
    type=click.Choice(["yaml", "json"]),
    default="json",
    required=False,
    help='Format output as YAML or JSON. Defaults to "json"',
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def arn_table(
    name: str,
    service: str,
    list_arn_types: bool,
    fmt: str = "json",
    verbose: str | None = None,
) -> None:
    """Query the ARN Table from the Policy Sentry database"""
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)
    query_arn_table(name, service, list_arn_types, fmt)


def query_arn_table(name: str, service: str, list_arn_types: bool, fmt: str) -> list[str] | dict[str, str]:
    """Query the ARN Table from the Policy Sentry database. Use this one when leveraging Policy Sentry as a library."""
    if LOCAL_DATASTORE_FILE_PATH.exists():
        logger.info(
            f"Using the Local IAM definition: {LOCAL_DATASTORE_FILE_PATH}. To leverage the bundled definition instead, remove the folder $HOME/.policy_sentry/"
        )
    else:
        # Otherwise, leverage the datastore inside the python package
        logger.debug("Leveraging the bundled IAM Definition.")
    # Get a list of all RAW ARN formats available through the service.
    if name is None and list_arn_types is False:
        output = get_raw_arns_for_service(service)
        print_list(output=output, fmt=fmt)
    # Get a list of all the ARN types per service, paired with the RAW ARNs
    elif name is None and list_arn_types:
        output = get_arn_types_for_service(service)
        print_dict(output=output, fmt=fmt)
    # Get the raw ARN format for the `cloud9` service with the short name
    # `environment`
    else:
        output = get_arn_type_details(service, name)
        print_dict(output=output, fmt=fmt)
    return output


@query.command(short_help="Query the condition table.")
@click.option(
    "--name",
    "-n",
    type=str,
    required=False,
    help="Get details on a specific condition key. Leave this blank to get a list of all condition keys "
    "available to the service.",
)
@click.option("--service", "-s", type=str, required=True, help="Filter according to AWS service.")
@click.option(
    "--fmt",
    type=click.Choice(["yaml", "json"]),
    default="json",
    required=False,
    help='Format output as YAML or JSON. Defaults to "json"',
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def condition_table(name: str, service: str, fmt: str, verbose: str | None) -> None:
    """Query the condition table from the Policy Sentry database"""
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)
    query_condition_table(name, service, fmt)


def query_condition_table(name: str, service: str, fmt: str = "json") -> list[str] | dict[str, str]:
    """Query the condition table from the Policy Sentry database.
    Use this one when leveraging Policy Sentry as a library."""
    if LOCAL_DATASTORE_FILE_PATH.exists():
        logger.info(
            f"Using the Local IAM definition: {LOCAL_DATASTORE_FILE_PATH}. To leverage the bundled definition instead, remove the folder $HOME/.policy_sentry/"
        )
    else:
        # Otherwise, leverage the datastore inside the python package
        logger.debug("Leveraging the bundled IAM Definition.")
    # Get a list of all condition keys available to the service
    if name is None:
        output = get_condition_keys_for_service(service)
        print_list(output=output, fmt=fmt)
    # Get details on the specific condition key
    else:
        output = get_condition_key_details(service, name)
        print_dict(output=output, fmt=fmt)
    return output


@query.command(short_help="Query the service table.")
@click.option(
    "--fmt",
    type=click.Choice(["yaml", "json", "csv"]),
    default="json",
    required=False,
    help='Format output as YAML, JSON or CSV. Defaults to "json"',
)
@click.option(
    "--verbose",
    "-v",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
def service_table(fmt: str, verbose: str | None) -> None:
    """Query the service table from the Policy Sentry database"""
    if verbose:
        log_level = getattr(logging, verbose.upper())
        set_stream_logger(level=log_level)
    query_service_table(fmt)


def query_service_table(fmt: str = "json") -> list[dict[str, str]]:
    """Query the service table from the Policy Sentry database.
    Use this one when leveraging Policy Sentry as a library."""
    if LOCAL_DATASTORE_FILE_PATH.exists():
        logger.info(
            f"Using the Local IAM definition: {LOCAL_DATASTORE_FILE_PATH}. To leverage the bundled definition instead, remove the folder $HOME/.policy_sentry/"
        )
    else:
        # Otherwise, leverage the datastore inside the python package
        logger.debug("Leveraging the bundled IAM Definition.")

    output = get_services_data()
    print_dict(output=output, fmt=fmt)
    return output



import unittest
from policy_sentry.util.conditions import is_condition_key_match


class ConditionsTestCase(unittest.TestCase):
    def test_is_condition_key_match_case_1(self):
        """exact match. return True"""
        document_key = "s3:prefix"
        str_to_match = "s3:prefix"
        self.assertTrue(is_condition_key_match(document_key, str_to_match))

    def test_is_condition_key_match_case_2(self):
        """no match. return False"""
        document_key = "s3:prefix"
        str_to_match = "secretsmanager:SecretId"
        self.assertFalse(is_condition_key_match(document_key, str_to_match))

    def test_is_condition_key_match_case_3(self):
        """match with string before $"""
        document_key = "aws:ResourceTag/${TagKey}"
        str_to_match = "aws:Resourcetag/${TagKey1}"
        self.assertTrue(is_condition_key_match(document_key, str_to_match))

    def test_is_condition_key_match_case_4(self):
        """match with string before <"""
        document_key = "s3:ExistingObjectTag/<key>"
        str_to_match = "s3:ExistingObjectTag/<sample-key>"
        self.assertTrue(is_condition_key_match(document_key, str_to_match))



"""
Functions that relate to manipulating files, loading files, and managing filepaths.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, cast

import yaml

logger = logging.getLogger(__name__)


def read_yaml_file(filename: str | Path) -> dict[str, Any]:
    """
    Reads a YAML file, safe loads, and returns the dictionary

    :param filename: name of the yaml file
    :return: dictionary of YAML file contents
    """
    with Path(filename).open(encoding="utf-8") as yaml_file:
        try:
            cfg = cast("dict[str, Any]", yaml.safe_load(yaml_file))
        except yaml.YAMLError as exc:
            logger.critical(exc)
    return cfg



import unittest
from policy_sentry.util.access_levels import determine_access_level_override
from policy_sentry.shared.awsdocs import get_action_access_level_overrides_from_yml


class AccessLevelOverride(unittest.TestCase):
    def test_passing_overall_iam_action_override(self):
        """test_passing_overall_iam_action_override: Tests iam:CreateAccessKey
        (in overrides file as Permissions management, but in the AWS docs as Write)"""
        desired_result = "Permissions management"
        action_overrides = get_action_access_level_overrides_from_yml("iam")
        result = determine_access_level_override("iam", "CreateAccessKey", "Write", action_overrides)
        self.assertEqual(result, desired_result)

    def test_overrides_yml_config(self):
        """test_overrides_yml_config: Tests the format of the overrides yml file for the RAM service"""
        desired_result = {
            "Permissions management": [
                "acceptresourceshareinvitation",
                "associateresourceshare",
                "createresourceshare",
                "deleteresourceshare",
                "disassociateresourceshare",
                "enablesharingwithawsorganization",
                "rejectresourceshareinvitation",
                "updateresourceshare",
            ],
            "Tagging": ["tagresource", "untagresource"],
        }
        ram_action_overrides = get_action_access_level_overrides_from_yml("ram")
        self.assertDictEqual(ram_action_overrides, desired_result)



"""
Methods that are basically text/string operations for specific purposes
"""


def capitalize_first_character(some_string: str) -> str:
    """
    Description: Capitalizes the first character of a string
    :param some_string:
    :return:
    """
    return " ".join("".join([w[0].upper(), w[1:].lower()]) for w in some_string.split())


def strip_special_characters(some_string: str) -> str:
    """Remove all special characters, punctuation, and spaces from a string"""
    # Input: "Special $#! characters   spaces 888323"
    # Output: 'Specialcharactersspaces888323'
    return "".join(e for e in some_string if e.isalnum())



import unittest

from policy_sentry.querying.arns import (
    get_raw_arns_for_service,
    get_arn_types_for_service,
    get_arn_type_details,
    get_resource_type_name_with_raw_arn,
    get_matching_raw_arns,
)


class QueryArnsTestCase(unittest.TestCase):
    def test_get_raw_arns_for_service(self):
        """querying.arns.get_raw_arns_for_service"""
        expected_results = [
            "arn:${Partition}:s3:${Region}:${Account}:accesspoint/${AccessPointName}",
            "arn:${Partition}:s3:::${BucketName}",
            "arn:${Partition}:s3:::${BucketName}/${ObjectName}",
            "arn:${Partition}:s3:${Region}:${Account}:job/${JobId}",
            "arn:${Partition}:s3:${Region}:${Account}:storage-lens/${ConfigId}",
        ]
        results = get_raw_arns_for_service("s3")
        self.maxDiff = None
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)

    def test_get_arn_types_for_service(self):
        """querying.arns.get_arn_types_for_service: Tests function that grabs arn_type and raw_arn pairs"""
        expected_results = {
            "accesspoint": "arn:${Partition}:s3:${Region}:${Account}:accesspoint/${AccessPointName}",
            "bucket": "arn:${Partition}:s3:::${BucketName}",
            "object": "arn:${Partition}:s3:::${BucketName}/${ObjectName}",
            "job": "arn:${Partition}:s3:${Region}:${Account}:job/${JobId}",
        }
        results = get_arn_types_for_service("s3")
        self.maxDiff = None
        for expected_result in expected_results:
            self.assertTrue(expected_result in results)

    def test_get_arn_type_details(self):
        """querying.arns.get_arn_type_details: Tests function that grabs details about a specific ARN name"""
        expected_results = {
            "resource_type_name": "environment",
            "raw_arn": "arn:${Partition}:cloud9:${Region}:${Account}:environment:${ResourceId}",
            "condition_keys": ["aws:ResourceTag/${TagKey}"],
        }
        results = get_arn_type_details("cloud9", "environment")
        # print(json.dumps(results, indent=4))
        self.assertEqual(results, expected_results)

    def test_get_resource_type_name_with_raw_arn(self):
        """querying.arns.get_resource_type_name_with_raw_arn"""
        raw_arn = "arn:${Partition}:cloud9:${Region}:${Account}:environment:${ResourceId}"
        self.assertTrue(get_resource_type_name_with_raw_arn(raw_arn), "environment")

    def test_get_matching_raw_arn(self):
        """querying.arns.get_matching_raw_arns"""
        self.assertEqual(
            get_matching_raw_arns("arn:aws:s3:::bucket_name"),
            ["arn:${Partition}:s3:::${BucketName}"],
        )
        self.assertEqual(
            get_matching_raw_arns("arn:aws:codecommit:us-east-1:123456789012:MyDemoRepo"),
            ["arn:${Partition}:codecommit:${Region}:${Account}:${RepositoryName}"],
        )
        self.assertEqual(
            get_matching_raw_arns("arn:aws:ssm:us-east-1:123456789012:parameter/test"),
            ["arn:${Partition}:ssm:${Region}:${Account}:parameter/${ParameterNameWithoutLeadingSlash}"],
        )
        self.assertEqual(
            get_matching_raw_arns("arn:aws:batch:region:account-id:job-definition/job-name:revision"),
            [
                "arn:${Partition}:batch:${Region}:${Account}:job-definition/${JobDefinitionName}",
                "arn:${Partition}:batch:${Region}:${Account}:job-definition/${JobDefinitionName}:${Revision}",
            ],
        )
        self.assertEqual(
            get_matching_raw_arns("arn:aws:states:region:account-id:stateMachine:stateMachineName"),
            [
                "arn:${Partition}:states:${Region}:${Account}:stateMachine:${StateMachineName}",
                "arn:${Partition}:states:${Region}:${Account}:stateMachine:${StateMachineName}:${StateMachineVersionId}",
                "arn:${Partition}:states:${Region}:${Account}:stateMachine:${StateMachineName}:${StateMachineAliasName}",
            ],
        )
        self.assertEqual(
            get_matching_raw_arns("arn:aws:states:region:account-id:execution:stateMachineName:executionName"),
            [
                "arn:${Partition}:states:${Region}:${Account}:execution:${StateMachineName}:${ExecutionId}",
                "arn:${Partition}:states:${Region}:${Account}:execution:${StateMachineName}/${MapRunLabel}:${ExecutionId}",
            ],
        )
        # self.assertEqual(get_matching_raw_arns("arn:aws:greengrass:region:account-id:/greengrass/definition/devices/1234567/versions/1"), ["arn:aws:greengrass:${Region}:${Account}:/greengrass/definition/devices/${DeviceDefinitionId}/versions/${VersionId}"])
        self.assertEqual(
            get_matching_raw_arns("arn:${Partition}:rds:region:account-id:db:mydatabase"),
            ["arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}"],
        )
        self.assertIn(
            "arn:${Partition}:rds:${Region}:${Account}:db:${DbInstanceName}",
            get_matching_raw_arns("arn:${Partition}:rds:region:account-id:*:*"),
        )
        self.assertEqual(
            get_matching_raw_arns("arn:${Partition}:rds:region:account-id:invalid-resource:*"),
            [],
        )
